# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for the lsl.sim.s60/tbw/tbn/drx modules."""

import os
import copy
import unittest
import tempfile
import numpy

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import tbw as tbwReader
from lsl.sim import tbw as tbwWriter
from lsl.reader import tbn as tbnReader
from lsl.sim import tbn as tbnWriter
from lsl.reader import drx as drxReader
from lsl.sim import drx as drxWriter
from lsl.sim import errors


__revision__ = "$Rev: 2470 $"
__version__  = "0.3"
__author__    = "Jayce Dowell"


tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')
drxFile = os.path.join(dataPath, 'tests', 'drx-test.dat')


class fake_TBW_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.tbw
    module."""
    
    testPath = None
    
    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""
        
        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-fakedata-', suffix='.tmp')
        
    def test_sim_frame(self):
        """Test the tbw.SimFrame class."""
        
        # Read in a TBW frame from the test file
        fh = open(tbwFile, 'rb')
        origFrame = tbwReader.readFrame(fh)
        fh.close()
        
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(origFrame)
        # Test the validity of the SimFrame
        self.assertTrue(fakeFrame.isValid())
        
    def test_write_frame(self):
        """Test that the TBW data writer works."""
        
        testFile = os.path.join(self.testPath, 'tbw-test-W.dat')
        
        nFrames = os.path.getsize(tbwFile) // tbwReader.FrameSize
        
        # Read in a TBW frame from the test file
        fh = open(tbwFile, 'rb')
        origFrames = []
        for i in xrange(nFrames):
            origFrames.append( tbwReader.readFrame(fh) )
        fh.close()
        
        # Write the data to a TBN test frame
        fh = open(testFile, 'wb')
        for origFrame in origFrames:
            rawFrame = tbwWriter.frame2frame(origFrame)
            rawFrame.tofile(fh)
        fh.close()
        
        # Read in the 
        fh = open(testFile, 'rb')
        fakeFrames = []
        for i in xrange(nFrames):
            fakeFrames.append( tbwReader.readFrame(fh) )
        fh.close()
        
        for fakeFrame,origFrame in zip(fakeFrames, origFrames):
            # Test values returned by info functions
            self.assertEqual(fakeFrame.parseID(), origFrame.parseID())
            self.assertEqual(fakeFrame.getDataBits(), origFrame.getDataBits())
            
            # Test raw header values
            self.assertTrue(fakeFrame.header.isTBW())
            self.assertEqual(fakeFrame.header.frameCount, origFrame.header.frameCount)
            self.assertEqual(fakeFrame.header.secondsCount, origFrame.header.secondsCount)
            
            # Test raw data values
            self.assertEqual(fakeFrame.data.timeTag, origFrame.data.timeTag)
            for i in range(400):
                self.assertEqual(fakeFrame.data.xy[0,i], origFrame.data.xy[0,i])
                self.assertEqual(fakeFrame.data.xy[1,i], origFrame.data.xy[1,i])
                
    def test_frame_data_errors(self):
        """Test the data error scenarios when validating a TBW SimFrame ."""
        
        # Read in a TBW frame from the test file
        fh = open(tbwFile, 'rb')
        origFrame = tbwReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong data type
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.xy = fakeFrame.data.xy.astype(numpy.complex64)
        self.assertRaises(errors.invalidDataType, fakeFrame.isValid, raiseErrors=True)
        
        # Try to validate frame with the wrong data size
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.xy = None
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.xy = fakeFrame.data.xy[0,:]
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.xy = fakeFrame.data.xy[:,0:50]
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.dataBits = 4
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        
    def test_frame_header_errors(self):
        """Test the header error scenarios when validating a TBW SimFrame."""
        
        # Read in a TBW frame from the test file
        fh = open(tbwFile, 'rb')
        origFrame = tbwReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong stand number
        fakeFrame = tbwWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.stand = 300
        self.assertRaises(errors.invalidStand, fakeFrame.isValid, raiseErrors=True)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""
        
        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class fake_TBN_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.tbn
    module."""
    
    testPath = None
    
    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""
        
        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-fakedata-', suffix='.tmp')
        
    def test_sim_frame(self):
        """Test the tbn.SimFrame class."""
        
        # Read in a TBN frame from the test file
        fh = open(tbnFile, 'rb')
        origFrame = tbnReader.readFrame(fh)
        fh.close()
        
        fakeFrame = tbnWriter.SimFrame()
        fakeFrame.loadFrame(origFrame)
        # Test the validity of the SimFrame
        self.assertTrue(fakeFrame.isValid())
        
    def test_write_frame(self):
        """Test that the TBN data writer works."""
        
        testFile = os.path.join(self.testPath, 'tbn-test-W.dat')
        
        nFrames = os.path.getsize(tbnFile) // tbnReader.FrameSize
        
        # Read in a TBN frame from the test file
        fh = open(tbnFile, 'rb')
        origFrames = []
        for i in xrange(nFrames):
            origFrames.append( tbnReader.readFrame(fh) )
        fh.close()
        
        # Write the data to a TBN test frame
        fh = open(testFile, 'wb')
        for origFrame in origFrames:
            rawFrame = tbnWriter.frame2frame(origFrame)
            rawFrame.tofile(fh)
        fh.close()
        
        # Read in the 
        fh = open(testFile, 'rb')
        fakeFrames = []
        for i in xrange(nFrames):
            fakeFrames.append( tbnReader.readFrame(fh) )
        fh.close()
        
        for fakeFrame,origFrame in zip(fakeFrames, origFrames):
            # Test values returned by info functions
            self.assertEqual(fakeFrame.parseID()[0], origFrame.parseID()[0])
            self.assertEqual(fakeFrame.parseID()[1], origFrame.parseID()[1])
            
            # Test raw header values
            self.assertTrue(fakeFrame.header.isTBN())
            self.assertEqual(fakeFrame.header.frameCount, origFrame.header.frameCount)
            self.assertEqual(fakeFrame.header.tuningWord, origFrame.header.tuningWord)
            
            # Test raw data values
            self.assertEqual(fakeFrame.data.timeTag, origFrame.data.timeTag)
            for i in range(512):
                self.assertEqual(fakeFrame.data.iq[i].real, origFrame.data.iq[i].real)
                self.assertEqual(fakeFrame.data.iq[i].imag, origFrame.data.iq[i].imag)
                
    def test_frame_data_errors(self):
        """Test the data error scenarios when validating a TBN SimFrame."""
        
        # Read in a TBN frame from the test file
        fh = open(tbnFile, 'rb')
        origFrame = tbnReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong data type
        fakeFrame = tbnWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = fakeFrame.data.iq.real
        self.assertRaises(errors.invalidDataType, fakeFrame.isValid, raiseErrors=True)
        
        # Try to validate frame with the wrong data size
        fakeFrame = tbnWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = None
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        fakeFrame = tbnWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = fakeFrame.data.iq[0:50]
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        
    def test_frame_header_errors(self):
        """Test the header error scenarios when validating a TBN SimFrame."""
        
        # Read in a TBN frame from the test file
        fh = open(tbnFile, 'rb')
        origFrame = tbnReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong stand number
        fakeFrame = tbnWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.stand = 300
        self.assertRaises(errors.invalidStand, fakeFrame.isValid, raiseErrors=True)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""
        
        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class fake_DRX_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.drx
    module."""
    
    testPath = None
    
    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""
        
        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-fakedata-', suffix='.tmp')
        
    def test_sim_frame(self):
        """Test the drx.SimFrame class."""
        
        # Read in a DRX frame from the test file
        fh = open(drxFile, 'rb')
        origFrame = drxReader.readFrame(fh)
        fh.close()
        
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(origFrame)
        # Test the validity of the SimFrame
        self.assertTrue(fakeFrame.isValid())
        
    def test_write_frame(self):
        """Test that the DRX data writer works."""
        
        testFile = os.path.join(self.testPath, 'drx-test-W.dat')
        
        nFrames = os.path.getsize(drxFile) // drxReader.FrameSize
        
        # Read in a TBN frame from the test file
        fh = open(drxFile, 'rb')
        origFrames = []
        for i in xrange(nFrames):
            origFrames.append( drxReader.readFrame(fh) )
        fh.close()
        
        # Write the data to a TBN test frame
        fh = open(testFile, 'wb')
        for origFrame in origFrames:
            rawFrame = drxWriter.frame2frame(origFrame)
            rawFrame.tofile(fh)
        fh.close()
        
        # Read in the 
        fh = open(testFile, 'rb')
        fakeFrames = []
        for i in xrange(nFrames):
            fakeFrames.append( drxReader.readFrame(fh) )
        fh.close()
        
        for fakeFrame,origFrame in zip(fakeFrames, origFrames):
            # Test values returned by info functions
            self.assertEqual(fakeFrame.parseID()[0], origFrame.parseID()[0])
            self.assertEqual(fakeFrame.parseID()[1], origFrame.parseID()[1])
            self.assertEqual(fakeFrame.parseID()[2], origFrame.parseID()[2])
            self.assertAlmostEqual(fakeFrame.getSampleRate(), origFrame.getSampleRate(), 4)
            
            # Test raw header values
            self.assertEqual(fakeFrame.header.secondsCount, origFrame.header.secondsCount)
            self.assertEqual(fakeFrame.header.decimation, origFrame.header.decimation)
            self.assertEqual(fakeFrame.header.timeOffset, origFrame.header.timeOffset)
            
            # Test raw data values
            self.assertEqual(fakeFrame.data.timeTag, origFrame.data.timeTag)
            self.assertEqual(fakeFrame.data.flags, origFrame.data.flags)
            for i in range(4096):
                self.assertEqual(fakeFrame.data.iq[i].real, origFrame.data.iq[i].real)
                self.assertEqual(fakeFrame.data.iq[i].imag, origFrame.data.iq[i].imag)
                
    def test_frame_data_errors(self):
        """Test the data error scenarios when validating a DRX SimFrame."""
        
        # Read in a DRX frame from the test file
        fh = open(drxFile, 'rb')
        origFrame = drxReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong data type
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = fakeFrame.data.iq.real
        self.assertRaises(errors.invalidDataType, fakeFrame.isValid, raiseErrors=True)
        
        # Try to validate frame with the wrong data size
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = None
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.iq = fakeFrame.data.iq[0:50]
        self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
        
    def test_frame_header_errors(self):
        """Test the header error scenarios when validating a DRX SimFrame."""
        
        # Read in a DRX frame from the test file
        fh = open(drxFile, 'rb')
        origFrame = drxReader.readFrame(fh)
        fh.close()
        
        # Try to validate frame with the wrong beam number
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.beam = 5
        self.assertRaises(errors.invalidBeam, fakeFrame.isValid, raiseErrors=True)
        
        # Try to validate frame with the wrong tuning number
        fakeFrame = drxWriter.SimFrame()
        fakeFrame.loadFrame(copy.deepcopy(origFrame))
        fakeFrame.tune = 3
        self.assertRaises(errors.invalidTune, fakeFrame.isValid, raiseErrors=True)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""
        
        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class fakedata_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(fake_TBW_tests))
        self.addTests(loader.loadTestsFromTestCase(fake_TBN_tests))
        self.addTests(loader.loadTestsFromTestCase(fake_DRX_tests))


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-

"""
Module that contains common values found in the ADP ICD.  The values 
are:
 * f_S - Sampling rate in samples per second
 * f_C - Width of each correlator frequency channel
 * T - Slot duration in seconds
 * T_2 - Sub-slot duration

Also included are two functions to convert between frequencies and ADP tuning 
words and functions for calculating the magnitude response of the TBN and DRX 
filters.

.. versionchanged:: 1.2.1
    Added in functions to calculate the magnitude response of the TBN and DRX
    filters
"""

import numpy
from scipy.signal import freqz, lfilter
from scipy.interpolate import interp1d


__version__ = '0.3'
__revision__ = '$Rev: 2644 $'
__all__ = ['fS', 'fC', 'T', 'T2', 'N_MAX', 'freq2word', 'word2freq', 'delaytoDPD', 'DPDtodelay', 
        'gaintoDPG', 'DPGtogain', 'tbnFilter', 'drxFilter', '__version__', '__revision__', '__all__']

fS = 196.0e6	# Hz
fC = 25e3		# Hz
T = 1.0		# seconds
T2 = 0.010	# seconds
N_MAX = 8192	# bytes

# FIR Filters
## TBN
_tbnFIR = [-3.22350e-05, -0.00021433,  0.0017756, -0.0044913,  0.0040327, 
            0.00735870,  -0.03218100,  0.0553980, -0.0398360, -0.0748920, 
            0.58308000,   0.58308000, -0.0748920, -0.0398360,  0.0553980,
        -0.03218100,   0.00735870,  0.0040327, -0.0044913,  0.0017756,
        -0.00021433,  -3.2235e-05]
## DRX
_drxFIR = [ 0.0111580, -0.0074330,  0.0085684, -0.0085984,  0.0070656, -0.0035905, 
        -0.0020837,  0.0099858, -0.0199800,  0.0316360, -0.0443470,  0.0573270, 
        -0.0696630,  0.0804420, -0.0888320,  0.0941650,  0.9040000,  0.0941650, 
        -0.0888320,  0.0804420, -0.0696630,  0.0573270, -0.0443470,  0.0316360, 
        -0.0199800,  0.0099858, -0.0020837, -0.0035905,  0.0070656, -0.0085984,  
            0.0085684, -0.0074330,  0.0111580]


_nPts = 1000 # Number of points to use in calculating the bandpasses


def freq2word(freq):
    """
    Given a frequency in Hz, convert it to the closest DP tuning word.
    """
    
    return int(round(freq*2**32 / fS))


def word2freq(word):
    """
    Given a DP tuning word, convert it to a frequncy in Hz.
    """
    
    return word*fS / 2**32


def delaytoDPD(delay):
    """
    Given a delay in ns, convert it to a course and fine portion and into the 
    final format expected by ADP (big endian 16.12 unsigned integer).
    """
    
    # Convert the delay to a combination of FIFO delays (~5.1 ns) and 
    # FIR delays (~0.3 ns)
    sample = int(round(delay * fS * 16 / 1e9))
    course = sample // 16
    fine   = sample % 16
    
    # Combine into one value
    combined = (course << 4) | fine
    
    # Convert to big-endian
    combined = ((combined & 0xFF) << 8) | ((combined >> 8) & 0xFF)
    
    return combined


def DPDtodelay(combined):
    """
    Given a delay value in the final format expect by ADP, return the delay in ns.
    """
    
    # Convert to little-endian
    combined = ((combined & 0xFF) << 8) | ((combined >> 8) & 0xFF)
    
    # Split
    fine = combined & 15
    course = (combined >> 4) & 4095
    
    # Convert to time
    delay = (course + fine/16.0) / fS
    delay *= 1e9
    
    return delay


def gaintoDPG(gain):
    """
    Given a gain (between 0 and 1), convert it to a gain in the final form 
    expected by ADP (big endian 16.1 signed integer).
    """
    
    # Convert
    combined = int(32767*gain)
    
    # Convert to big-endian
    combined = ((combined & 0xFF) << 8) | ((combined >> 8) & 0xFF)
    
    return combined


def DPGtogain(combined):
    """
    Given a gain value in the final format expected by ADP, return the gain
    as a decimal value (0 to 1).
    """
    
    # Convert to little-endian
    combined = ((combined & 0xFF) << 8) | ((combined >> 8) & 0xFF)
    
    # Convert back
    gain = combined / 32767.0
    
    return gain


def tbnFilter(sampleRate=1e5, nPts=_nPts):
    """
    Return a function that will generate the shape of a TBN filter for a given sample
    rate.
    """
    
    # Part 0 - FIR filter
    h, wFIR = freqz(_tbnFIR, 1, nPts)
    w = numpy.abs(wFIR)
    
    # Convert to a "real" frequency and magnitude response
    h *= sampleRate / 2.0 / numpy.pi
    w = numpy.abs(w)**2
    
    # Mirror
    h = numpy.concatenate([-h[::-1], h[1:]])
    w = numpy.concatenate([ w[::-1], w[1:]])
    
    # Return the interpolating function
    return interp1d(h, w/w.max(), kind='cubic', bounds_error=False, fill_value=0.0)


def drxFilter(sampleRate=9.8e6, nPts=_nPts):
    """
    Return a function that will generate the shape of a DRX filter for a given sample
    rate.
    
    Based on memo DRX0001.
    """
    # Part 0 - FIR filter
    h, wFIR = freqz(_drxFIR, 1, nPts)
    w = numpy.abs(wFIR)
    
    # Convert to a "real" frequency and magnitude response
    h *= sampleRate / 2.0 / numpy.pi
    w = numpy.abs(w)**2
    
    # Mirror
    h = numpy.concatenate([-h[::-1], h[1:]])
    w = numpy.concatenate([w[::-1], w[1:]])
    
    # Return the interpolating function
    return interp1d(h, w/w.max(), kind='cubic', bounds_error=False, fill_value=0.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example script for reading in DRX data and writing it to a TS-FITS file."""

import sys
import time
import ephem

from lsl.reader.ldp import LWA1DataFile
from lsl.writer import tsfits
from lsl.astro import unix_to_utcjd, DJD_OFFSET


def main(args):
    idf = LWA1DataFile(args[0])
    nFramesFile = idf.getInfo('nFrames')
    
    srate = idf.getInfo('sampleRate')
    beam = idf.getInfo('beam')
    beampols = idf.getInfo('beampols')
    
    # Date
    beginDate = ephem.Date(unix_to_utcjd(idf.getInfo('tStart')) - DJD_OFFSET)
    
    # File summary
    print "Filename: %s" % args[0]
    print "Date of First Frame: %s" % str(beginDate)
    print "Beam: %i" % beam
    print "Tune/Pols: %ii" % beampols
    print "Sample Rate: %i Hz" % srate
    print "Frames: %i (%.3f s)" % (nFramesFile, 1.0 * nFramesFile / beampols * 4096 / srate)
    print "---"

    tStart = time.time()
    
    # Create a new FITS file with the name 'drx-tsfits.fits'
    fitsFile = tsfits.TBN('drx-tsfits.fits')
    
    nSamples = 3400
    
    count = {}
    masterCount = 0
    for i in xrange(nSamples):
        for j in xrange(beampols):
            frame = idf.readFrame()
            beam, pol, tune = frame.parseID()
            try:
                count[beam] += 1
            except KeyError:
                count[beam] = 1
                
            fitsFile.addStandData(frame)
            
        masterCount = masterCount + 1
        
    tEnd = time.time()
    print 'Read %i frames in %0.3f s (%0.1f frames/s)' % (beampols*nSamples, (tEnd-tStart), beampols*nSamples/(tEnd-tStart))
    
    idf.close()
    fitsFile.close()
    fitsFile.info()
    
    # Summary information about the file that was just read in
    print "Summary:"
    for beam in sorted(count.keys()):
        print "Beam: %2i, Frames: %5i" % (beam, count[beam])


if __name__ == "__main__":
    main(sys.argv[1:])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Predict driftcurve for a given site using a given antenna model."""

import os
import sys
import math
import numpy
import pylab
import argparse

from scipy.interpolate import interp1d

from lsl import skymap, astro
from lsl.common import stations
from lsl.common.paths import data as dataPath
from lsl.misc import parser as aph

__revision__ = "$Revision: 94 $"
__version__  = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


def main(args):
    # Validate
    if args.pol not in ('EW', 'NS'):
        raise ValueError("Invalid polarization: %s" % args.pol)
        
    # Get the site information
    if args.lwasv:
        nam = 'lwasv'
        sta = stations.lwasv
    elif args.ovro_lwa:
        nam = 'ovro'
        sta = stations.lwa1
        sta.lat, sta.lon, sta.elev = ('37.2397808', '-118.2816819', 1183.4839)
    else:
        nam = 'lwa1'
        sta = stations.lwa1
        
    # Read in the skymap (GSM or LF map @ 74 MHz)
    if not args.lfsm:
        smap = skymap.SkyMapGSM(freqMHz=args.frequency/1e6)
        if args.verbose:
            print "Read in GSM map at %.2f MHz of %s pixels; min=%f, max=%f" % (args.frequency/1e6, len(smap.ra), smap._power.min(), smap._power.max())
    else:
        smap = skymap.SkyMapLFSM(freqMHz=args.frequency/1e6)
        if args.verbose:
            print "Read in LFSM map at %.2f MHz of %s pixels; min=%f, max=%f" % (args.frequency/1e6, len(smap.ra), smap._power.min(), smap._power.max())
    
    # Get the emperical model of the beam and compute it for the correct frequencies
    beamDict = numpy.load(os.path.join(dataPath, 'lwa1-dipole-emp.npz'))
    if args.pol == 'EW':
        beamCoeff = beamDict['fitX']
    else:
        beamCoeff = beamDict['fitY']
    try:
        beamDict.close()
    except AttributeError:
        pass
    alphaE = numpy.polyval(beamCoeff[0,0,:], args.frequency)
    betaE =  numpy.polyval(beamCoeff[0,1,:], args.frequency)
    gammaE = numpy.polyval(beamCoeff[0,2,:], args.frequency)
    deltaE = numpy.polyval(beamCoeff[0,3,:], args.frequency)
    alphaH = numpy.polyval(beamCoeff[1,0,:], args.frequency)
    betaH =  numpy.polyval(beamCoeff[1,1,:], args.frequency)
    gammaH = numpy.polyval(beamCoeff[1,2,:], args.frequency)
    deltaH = numpy.polyval(beamCoeff[1,3,:], args.frequency)
    if args.verbose:
        print "Beam Coeffs. X: a=%.2f, b=%.2f, g=%.2f, d=%.2f" % (alphaH, betaH, gammaH, deltaH)
        print "Beam Coeffs. Y: a=%.2f, b=%.2f, g=%.2f, d=%.2f" % (alphaE, betaE, gammaE, deltaE)
        
    if args.empirical:
        corrDict = numpy.load(os.path.join(dataPath, 'lwa1-dipole-cor.npz'))
        cFreqs = corrDict['freqs']
        cAlts  = corrDict['alts']
        if corrDict['degrees'].item():
            cAlts *= numpy.pi / 180.0
        cCorrs = corrDict['corrs']
        corrDict.close()
        
        if args.frequency/1e6 < cFreqs.min() or args.frequency/1e6 > cFreqs.max():
            print "WARNING: Input frequency of %.3f MHz is out of range, skipping correction" % (args.frequency/1e6,)
            corrFnc = None
        else:
            fCors = cAlts*0.0
            for i in xrange(fCors.size):
                ffnc = interp1d(cFreqs, cCorrs[:,i], bounds_error=False)
                fCors[i] = ffnc(args.frequency/1e6)
            corrFnc = interp1d(cAlts, fCors, bounds_error=False)
            
    else:
        corrFnc = None
        
    def BeamPattern(az, alt, corr=corrFnc):
        zaR = numpy.pi/2 - alt*numpy.pi / 180.0 
        azR = az*numpy.pi / 180.0
        
        c = 1.0
        if corrFnc is not None:
            c = corrFnc(alt*numpy.pi / 180.0)
            c = numpy.where(numpy.isfinite(c), c, 1.0)
            
        pE = (1-(2*zaR/numpy.pi)**alphaE)*numpy.cos(zaR)**betaE + gammaE*(2*zaR/numpy.pi)*numpy.cos(zaR)**deltaE
        pH = (1-(2*zaR/numpy.pi)**alphaH)*numpy.cos(zaR)**betaH + gammaH*(2*zaR/numpy.pi)*numpy.cos(zaR)**deltaH

        return c*numpy.sqrt((pE*numpy.cos(azR))**2 + (pH*numpy.sin(azR))**2)

    if args.do_plot:
        az = numpy.zeros((90,360))
        alt = numpy.zeros((90,360))
        for i in range(360):
            az[:,i] = i
        for i in range(90):
            alt[i,:] = i
        pylab.figure(1)
        pylab.title("Beam Response: %s pol. @ %0.2f MHz" % (args.pol, args.frequency/1e6))
        pylab.imshow(BeamPattern(az, alt), extent=(0,359, 0,89), origin='lower')
        pylab.xlabel("Azimuth [deg]")
        pylab.ylabel("Altitude [deg]")
        pylab.grid(1)
        pylab.draw()
    
    # Calculate times in both site LST and UTC
    t0 = astro.get_julian_from_sys()
    lst = astro.get_local_sidereal_time(sta.long*180.0/math.pi, t0) / 24.0
    t0 -= lst*(23.933/24.0) # Compensate for shorter sidereal days
    times = numpy.arange(0.0, 1.0, args.time_step/1440.0) + t0
    
    lstList = []
    powListAnt = [] 
    
    for t in times:
        # Project skymap to site location and observation time
        pmap = skymap.ProjectedSkyMap(smap, sta.lat*180.0/math.pi, sta.long*180.0/math.pi, t)
        lst = astro.get_local_sidereal_time(sta.long*180.0/math.pi, t)
        lstList.append(lst)
        
        # Convolution of user antenna pattern with visible skymap
        gain = BeamPattern(pmap.visibleAz, pmap.visibleAlt)
        powerAnt = (pmap.visiblePower * gain).sum() / gain.sum()
        powListAnt.append(powerAnt)

        if args.verbose:
            lstH = int(lst)
            lstM = int((lst - lstH)*60.0)
            lstS = ((lst - lstH)*60.0 - lstM)*60.0
            sys.stdout.write("LST: %02i:%02i:%04.1f, Power_ant: %.1f K\r" % (lstH, lstM, lstS, powerAnt))
            sys.stdout.flush()
    sys.stdout.write("\n")
            
    # plot results
    if args.do_plot:
        pylab.figure(2)
        pylab.title("Driftcurve: %s pol. @ %0.2f MHz - %s" % \
            (args.pol, args.frequency/1e6, nam.upper()))
        pylab.plot(lstList, powListAnt, "ro", label="Antenna Pattern")
        pylab.xlabel("LST [hours]")
        pylab.ylabel("Temp. [K]")
        pylab.grid(2)
        pylab.draw()
        pylab.show()
    
    outputFile = "driftcurve_%s_%s_%.2f.txt" % (nam, args.pol, args.frequency/1e6)
    print "Writing driftcurve to file '%s'" % outputFile
    mf = file(outputFile, "w")
    for lst,pow in zip(lstList, powListAnt):
        mf.write("%f  %f\n" % (lst, pow))
    mf.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='simulate a drift curve for a dipole at LWA1 observing at a given frequency in MHz', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('-f', '--frequency', type=aph.frequency, default='74.0', 
                    help='frequency of the simulation in MHz')
    sgroup = parser.add_mutually_exclusive_group(required=False)
    sgroup.add_argument('-s', '--lwasv', action='store_true', 
                        help='calculate for LWA-SV instead of LWA1')
    sgroup.add_argument('-o', '--ovro-lwa', action='store_true', 
                        help='calculate for OVRO-LWA instead of LWA1')
    parser.add_argument('-p', '--pol', type=str, default='EW', 
                        help='polarization of the simulations (NS or EW)')
    parser.add_argument('-e', '--empirical', action='store_true', 
                    help='enable empirical corrections to the dipole model (valid from 35 to 80 MHz)')
    parser.add_argument('-l', '--lfsm', action='store_true', 
                        help='use LFSM instead of GSM')
    parser.add_argument('-t', '--time-step', type=float, default=10.0, 
                        help='time step of the simulation in minutes')
    parser.add_argument('-x', '--do-plot', action='store_true', 
                        help='plot the driftcurve data')
    parser.add_argument('-v', '--verbose', action='store_true', 
                        help='run %(prog)s in verbose mode')
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import sys
import urllib
import hashlib
import argparse
from datetime import datetime

from lsl.common.paths import data as dataPath


# Regular expression for finding the SSMIF version from the file contents
versionRE = re.compile(r'(?P<year>\d{4}) (?P<month>[a-zA-Z]{3,4}) (?P<day>\d{1,2}) by (?P<author>.*)')


def parseIndex(index):
    """
    Parse the archive listing of SSMIF version and return a list of 
    filename/date tuples.
    """
    
    from xml.etree import ElementTree as ET
    from BeautifulSoup import BeautifulSoup
    
    # Find the table
    start = index.find('<table>')
    stop  = index.find('</table>')
    index = index[start:stop+8]
    
    # Clean it up in such a way that ElementTree can parse it
    myMassage = [(re.compile('<!([^--].*)>'), lambda match: '<!--' + match.group(1) + '-->'), 
            (re.compile('<hr>'), lambda match: ''), 
            (re.compile('&nbsp;'), lambda match: ' '), 
            (re.compile('<a.*?>(.*)</a>'), lambda mtch: mtch.group(1))]
    soup = BeautifulSoup(index, markupMassage=myMassage)
    index = soup.prettify()
    
    # Parse it
    table = ET.XML(index)
    rows = iter(table)
    
    # Extract the SSMIF entries
    versions = []
    for row in rows:
        values = [col.text for col in row]
        if len(values) != 5:
            continue
            
        filename = values[1].lstrip().rstrip()
        if filename[:5] != 'SSMIF':
            continue
        if filename.find('CURRENT') != -1:
            continue
            
        date = filename.split('_', 1)[1]
        date = date.split('.', 1)[0]
        versions.append( (filename, date) )
        
    # Done
    return versions


def getMD5(filename, blockSize=262144):
    """
    Compute the MD5 checksum of a file.
    """
    
    fh = open(filename, 'r')
    
    m = hashlib.md5()
    while True:
        block = fh.read(blockSize)
        if len(block) == 0:
            break
        m.update(block)
        
    fh.close()
    
    return m.hexdigest()


def main(args):
    # Current LSL SSMIF
    if args.lwasv:
        _name = 'LWA-SV'
        _ssmif = os.path.join(dataPath, 'lwa1-ssmif.txt')
        _url = "https://lda10g.alliance.unm.edu/metadata/lwasv/ssmif/"
    else:
        _name = 'LWA1'
        _ssmif = os.path.join(dataPath, 'lwa1-ssmif.txt')
        _url = "https://lda10g.alliance.unm.edu/metadata/lwa1/ssmif/"
        
    urlToDownload = None
    if args.revert:
        # Revert and Upgrade URL
        
        try:
            ## Retrieve the list
            ah = urllib.urlopen(_url)
            index = ah.read()
            ah.close()
            
            ## Parse
            versions = parseIndex(index)
            
            ## Prompt the user for the version to revert to
            for i,(filename,date) in enumerate(versions):
                print "%i: %s" % (i, filename)
            i = -1
            while i not in range(0, len(versions)):
                i = raw_input("Enter SSMIF to revert to: ")
                try:
                    i = int(i)
                except ValueError:
                    print "-> Invalid value"
                    i = -1
            print " "
            
            ## Build the URL
            urlToDownload = "%s/%s" % (_url, versions[i][0])
        except Exception, e:
            print "Error:  Cannot process reversion, %s" % str(e)
            
    elif args.update:
        # Update to the latest version
        
        urlToDownload = "%s/SSMIF_CURRENT.txt" % _url
        
    elif args.file is not None:
        # Use the specified file
        
        urlToDownload = os.path.abspath(args.file)
        
    # Put the new SSMIF in place
    if urlToDownload is not None:
        ## Retrieve
        try:
            ah = urllib.urlopen(urlToDownload)
            newSSMIF = ah.read()
            ah.close()
        except Exception, e:
            print "Error:  Cannot download SSMIF, %s" % str(e)
            
        ## Save
        try:
            fh = open(_ssmif, 'wb')
            fh.write(newSSMIF)
            fh.close()
        except Exception, e:
            print "Error:  Cannot %s SSMIF, %s" % ('update' if args.update else 'revert', str(e))
            
    # Summarize the SSMIF
    ## Filesystem information
    size = os.path.getsize(_ssmif)
    mtime = datetime.utcfromtimestamp(os.stat(_ssmif)[8])
    age = datetime.utcnow() - mtime
    
    ## MD5 checksum
    md5 = getMD5(_ssmif)
    
    ## SSMIF version (date)
    fh = open(_ssmif, 'r')
    lines = [fh.readline() for i in xrange(10)]
    fh.close()
    
    version = None
    for line in lines:
        mtch = versionRE.search(line)
        if mtch is not None:
            try:
                version = datetime.strptime("%s %s %s" % (mtch.group('year'), mtch.group('month'), mtch.group('day')), "%Y %b %d")
            except:
                version = datetime.strptime("%s %s %s" % (mtch.group('year'), mtch.group('month'), mtch.group('day')), "%Y %B %d")
            break
            
    print "LSL %s SSMIF%s:" % (_name.upper(), ' (updated)' if args.update else '',)
    print "  Size: %i bytes" % size
    print "  SSMIF Version: %s" % version.strftime("%Y %b %d")
    print "  File Last Modified: %s (%i day%s ago)" % (mtime.strftime("%Y-%m-%d %H:%M:%S UTC"), age.days, 's' if age.days != 1 else '')
    print "  MD5 Sum: %s" % md5


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description='update the internal LWA station SSMIFs used by LSL', 
            epilog='The -s/--lwasv option changes the behavior of all other options, e.g., -u/--update updates the LWA-SV SSMIF.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
            )
    parser.add_argument('-s', '--lwasv', action='store_true', 
                        help='update LWA-SV instead of LWA1')
    parser.add_argument('-u', '--update', action='store_true', 
                        help='update the default LWA1 SSMIF')
    parser.add_argument('-r', '--revert', action='store_true', 
                        help='reveret the default LWA1 SSMIF to an older version')
    parser.add_argument('-f', '--file', type=str, 
                        help='update the default LWA1 SSMIF using the specified file')
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-

"""
Python module to reading in data from both 12-bit and 4-bit TBW files.  
This module defines the following classes for storing the TBW data found in
a file:

Frame
  object that contains all data associated with a particular TBW frame.  The 
  primary consituents of each frame are:
    * FrameHeader - the TBW frame header object and
    * FrameData   - the TBW frame data object.  
Combined, these two objects contain all of the information found in the 
original TBW frame.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.  readFrame 
is designed to work with both 4-bit and 12-bit observations.

For describing the format of data in the file, two function are provided:

getDataBits
  read in the first frame of an open file handle and return whether or not 
  the data is 12 or 4-bit

getFramesPerObs
  read in the first several frames to see how many stands are found in the 
  data.
.. note::
    This function is a little flaky on TBW data sets that have less 
    than a full complement or 12M (36M) samples.

.. versionchanged:: 0.4.0
    Switched over from pure Python readers to the new C-base Go Fast! readers.
"""

import copy
import numpy

from lsl.common import dp as dp_common
from lsl.reader._gofast import readTBW
from lsl.reader._gofast import syncError as gsyncError
from lsl.reader._gofast import eofError as geofError
from lsl.reader.errors import syncError, eofError

__version__ = '0.6'
__revision__ = '$Rev: 2391 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readFrame', 
           'FrameSize', 'getDataBits', 'getFramesPerObs', 
           '__version__', '__revision__', '__all__']

FrameSize = 1224


class FrameHeader(object):
    """
    Class that stores the information found in the header of a TBW 
    frame.  All three fields listed in the DP ICD version H are stored as 
    well as the original binary header data.
    """

    def __init__(self, frameCount=None, secondsCount=None, tbwID=None):
        self.frameCount = frameCount
        self.secondsCount = secondsCount
        self.tbwID = tbwID

    def isTBW(self):
        """
        Function to check if the data is really TBW and not TBN by examining
        the TBW ID field.  Returns True if the data is TBW, false otherwise.
        """

        mode = (self.tbwID>>15)&1
        if mode == 1:
            return True
        else:
            return False

    def parseID(self):
        """
        Function to parse the TBW ID field and return the stand number.
        """

        # Why &1023?  Well, from DP ICD revision H, it seems that the stand count 
        # only goes up 260.  So, channel numbers should range up to 520, which can
        # be represented as 10 bits or 1023.
        stand = self.tbwID&1023

        return stand

    def getDataBits(self):
        """
        Function to parse the TBW ID field and return the size of number of 
        bits that comprise the data.  12 is returned for 12-bit data, and 4 
        for 4-bit data.
        """

        bits = (self.tbwID>>14)&1
        if bits == 0:
            dataBits = 12
        else:
            dataBits = 4

        return dataBits


class FrameData(object):
    """
    Class that stores the information found in the data section of a TBW
    frame.  Both fields listed in the DP ICD version H are stored.
    """

    def __init__(self, timeTag=None, samples=400, xy=None):
        self.timeTag = timeTag
        self.xy = xy

    def getTime(self):
        """
        Function to convert the time tag from samples since the UNIX epoch
        (UTC 1970-01-01 00:00:00) to seconds since the UNIX epoch.
        """

        seconds = self.timeTag / dp_common.fS
        
        return seconds


class Frame(object):
    """
    Class that stores the information contained within a single TBW 
    frame.  It's properties are FrameHeader and FrameData objects.
    """

    def __init__(self, header=None, data=None):
        if header is None:
            self.header = FrameHeader()
        else:
            self.header = header
            
        if data is None:
            self.data = FrameData()
        else:
            self.data = data
            
        self.valid = True
        
    def isTBW(self):
        """
        Convenience wrapper for the Frame.FrameHeader.isTBW function.
        """
        
        return self.header.isTBW()
        
    def parseID(self):
        """
        Convenience wrapper for the Frame.FrameHeader.parseID 
        function.
        """
        
        return self.header.parseID()

    def getDataBits(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getDataBits 
        function.
        """
        
        return self.header.getDataBits()

    def getTime(self):
        """
        Convenience wrapper for the Frame.FrameData.getTime function.
        """
        
        return self.data.getTime()
            
    def __add__(self, y):
        """
        Add the data sections of two frames together or add a number 
        to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame += y
        return newFrame
    
    def __iadd__(self, y):
        """
        In-place add the data sections of two frames together or add 
        a number to every element in the data section.
        """
        
        try:
            self.data.xy += y.data.xy
        except AttributeError:
            self.data.xy += numpy.int16(y)
        return self
        
    def __mul__(self, y):
        """
        Multiple the data sections of two frames together or multiply 
        a number to every element in the data section.
        """

        newFrame = copy.deepcopy(self)
        newFrame *= y
        return newFrame
            
    def __imul__(self, y):
        """
        In-place multiple the data sections of two frames together or 
        multiply a number to every element in the data section.
        """
        
        try:
            self.data.xy *= y.data.xy
        except AttributeError:
            self.data.xy *= numpy.int16(y)
        return self
            
    def __eq__(self, y):
        """
        Check if the time tags of two frames are equal or if the time
        tag is equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX == tY:
            return True
        else:
            return False
            
    def __ne__(self, y):
        """
        Check if the time tags of two frames are not equal or if the time
        tag is not equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX != tY:
            return True
        else:
            return False
            
    def __gt__(self, y):
        """
        Check if the time tag of the first frame is greater than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX > tY:
            return True
        else:
            return False
            
    def __ge__(self, y):
        """
        Check if the time tag of the first frame is greater than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX >= tY:
            return True
        else:
            return False
            
    def __lt__(self, y):
        """
        Check if the time tag of the first frame is less than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX < tY:
            return True
        else:
            return False
            
    def __le__(self, y):
        """
        Check if the time tag of the first frame is less than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX <= tY:
            return True
        else:
            return False
            
    def __cmp__(self, y):
        """
        Compare two frames based on the time tags.  This is helpful for 
        sorting things.
        """
        
        tX = self.data.timeTag
        tY = y.data.timeTag
        if tY > tX:
            return -1
        elif tX > tY:
            return 1
        else:
            return 0


def readFrame(filehandle, Verbose=False):
    """
    Function to read in a single TBW frame (header+data) and store the 
    contents as a Frame object.  This function wraps readerHeader and 
    readData[(12)|4].
    """
    
    # New Go Fast! (TM) method
    try:
        newFrame = readTBW(filehandle, Frame())
    except gsyncError:
        mark = filehandle.tell() - FrameSize
        raise syncError(location=mark)
    except geofError:
        raise eofError

    return newFrame


def getDataBits(filehandle):
    """
    Find out the number of data bits used in the file be reading in the 
    first frame.
    """

    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()

    # Read a frame
    cFrame = readFrame(filehandle)

    # Get the number of bits used to represent the data
    dataBits = cFrame.getDataBits()

    # Return to the place in the file where we started
    filehandle.seek(fhStart)

    return dataBits


def getFramesPerObs(filehandle):
    """
    Find out how many frames are present per observation by examining 
    the first frames for what would be 260 stands.  This is done by reading
    two frames and then skipping the next 30,000.
    
    .. note::
        Post-IOC it is probably simpler to adopt a value of the number of 
        frames per observation of 260 rather than try to find it from the
        file.
    """

    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()

    idCodes = []
    for i in range(260):
        currentPosition = filehandle.tell()
        try:
            cFrame1 = readFrame(filehandle)
            cFrame2 = readFrame(filehandle)
        except eofError:
            break
        except syncError:
            continue

        cID = cFrame1.parseID()
        if cID not in idCodes:
            idCodes.append(cID)
        cID = cFrame2.parseID()
        if cID not in idCodes:
            idCodes.append(cID)

        # Junk 30,000 frames since that is how many frames there are per stand
        filehandle.seek(currentPosition+30000*FrameSize)

    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Get the length of the stand list and return
    return len(idCodes)



# -*- coding: utf-8 -*

"""
This module stores various functions that are needed for computing UV 
coverage and time delays.  The functions in the module:
  * compute the u, v, and w coordinates of all baselines defined by an array 
    of stands
  * compute the track through the uv-plane of a collection of baselines as 
    the Earth rotates.
    
.. versionchanged:: 0.4.0
    Removed function for dealing with meta-data (position, cable length, etc.) 
    for individual stands since these are wrapped in the new :mod:`lsl.common.stations`
    module.
    
.. versionchanged:: 1.0.0
    Generalized the computeUVW() and computeUVTrack() functions.
"""

import numpy

from lsl.common.stations import lwa1
from lsl.common.constants import deg2rad, c as speedOfLight

__version__ = '0.6'
__revision__ = '$Rev: 2390 $'
__all__ = ['getBaselines', 'baseline2antenna', 'antenna2baseline', 'computeUVW', 'computeUVTrack', 
           'uvUtilsError', '__version__', '__revision__', '__all__']


class uvUtilsError(Exception):
    """Base exception class for this module"""
    
    def __init__(self, strerror, *args, **kwds):
        self.strerror = strerror
        super(uvUtilsError, self).__init__(*args, **kwds)
        
    def __str__(self):
        return "%s" % self.strerror


def getBaselines(antennas, antennas2=None, IncludeAuto=False, Indicies=False):
    """
    Generate a list of two-element tuples that describe which antennae
    compose each of the output uvw triplets from computeUVW/computeUVTrack.
    If the Indicies keyword is set to True, the two-element tuples 
    contain the indicies of the stands array used, rather than the actual
    stand numbers.
    """

    if IncludeAuto:
        offset = 0
    else:
        offset = 1

    out = []
    N = len(antennas)
    
    # If we don't have an antennas2 array, use antennas again
    if antennas2 is None:
        antennas2 = antennas
    
    for i in range(0, N-offset):
        for j in range(i+offset, N):
            if Indicies:
                out.append( (i, j) )
            else:
                out.append( (antennas[i], antennas2[j]) )

    return out
    

def baseline2antenna(baseline, antennas, antennas2=None, BaselineList=None, IncludeAuto=False, Indicies=False):
    """
    Given a baseline number, a list of stands, and options of how the base-
    line listed was generated, convert the baseline number to antenna numbers.
    Alternatively, use a list of baselines instead of generating a new list.  
    This utility is useful for figuring out what antennae comprise a baseline.
    """
    
    # If we don't have an antennas2 array, use antennas again
    if antennas2 is None:
        antennas2 = antennas

    # Build up the list of baselines using the options provided
    if BaselineList is None:
        BaselineList = getBaselines(antennas, antennas2=antennas2, IncludeAuto=IncludeAuto, Indicies=Indicies)
    
    # Select the correct one and return based on the value of Indicies
    i,j = BaselineList[baseline]
    return i,j


def antenna2baseline(ant1, ant2, antennas, antennas2=None, BaselineList=None, IncludeAuto=False, Indicies=False):
    """
    Given two antenna numbers, a list of stands, and options to how the base-
    line listed was generated, convert the antenna pair to  a baseline number. 
    This utility is useful for picking out a particular pair from a list of
    baselines.
    """
    
    # If we don't have an antennas2 array, use antennas again
    if antennas2 is None:
        antennas2 = antennas

    # Build up the list of baselines using the options provided
    if BaselineList is None:
        BaselineList = getBaselines(antennas, antennas2=antennas2, IncludeAuto=IncludeAuto, Indicies=Indicies)
    
    # Loop over the baselines until we find one that matches.  If we don't find 
    # one, return -1
    i = 0
    for baseline in BaselineList:
        if ant1 in baseline and ant2 in baseline:
            return i
        else:
            i = i + 1
    else:
        return -1


def computeUVW(antennas, HA=0.0, dec=34.070, freq=49.0e6, site=lwa1, IncludeAuto=False):
    """
    Compute the uvw converate of a baselines formed by a collection of 
    stands.  The coverage is computed at a given HA (in hours) and 
    declination (in degrees) for a given site.  The frequency provided 
    (in Hz) can either as a scalar or as a numpy array.  If more than one 
    frequency is given, the output is a three dimensional with dimensions 
    of baselines, uvw, and frequencies.
    
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.station` module instead of a list of stand ID
        numbers.
        
    .. versionchanged:: 1.0.0
        Added a keyword (site) to specify the station used for the 
        observation.
        
    .. versionchanged:: 1.1.2
        Updated to work with lists in a transparent manner.
    """
    
    # Try this so that freq can be either a scalar, a list, or an array
    try: 
        freq.size
        assert(freq.shape != ())
    except (AttributeError, AssertionError):
        freq = numpy.array(freq, ndmin=1)
        
    baselines = getBaselines(antennas, IncludeAuto=IncludeAuto, Indicies=True)
    Nbase = len(baselines)
    Nfreq = freq.size
    uvw = numpy.zeros((Nbase,3,Nfreq))

    # Phase center coordinates
    # Convert numbers to radians and, for HA, hours to degrees
    HA2 = HA * 15.0 * deg2rad
    dec2 = dec * deg2rad
    lat2 = site.lat
    
    # Coordinate transformation matrices
    trans1 = numpy.matrix([[0, -numpy.sin(lat2), numpy.cos(lat2)],
                    [1,  0,               0],
                    [0,  numpy.cos(lat2), numpy.sin(lat2)]])
    trans2 = numpy.matrix([[ numpy.sin(HA2),                  numpy.cos(HA2),                 0],
                    [-numpy.sin(dec2)*numpy.cos(HA2),  numpy.sin(dec2)*numpy.sin(HA2), numpy.cos(dec2)],
                    [ numpy.cos(dec2)*numpy.cos(HA2), -numpy.cos(dec2)*numpy.sin(HA2), numpy.sin(dec2)]])
                    
    for k,(i,j) in enumerate(baselines):
        # Go from a east, north, up coordinate system to a celestial equation, 
        # east, north celestial pole system
        xyzPrime = antennas[i].stand - antennas[j].stand
        xyz = trans1*numpy.matrix([[xyzPrime[0]], [xyzPrime[1]], [xyzPrime[2]]])
        
        # Go from CE, east, NCP to u, v, w
        temp = trans2*xyz
        uvw[k,:,:] = temp[:,0] * freq.ravel() / speedOfLight
        
    uvw.shape = (Nbase,3)+freq.shape
    
    return uvw


def computeUVTrack(antennas, dec=34.070, freq=49.0e6, site=lwa1):
    """
    Whereas computeUVW provides the uvw coverage at a particular time, 
    computeUVTrack provides the complete uv plane track for a long 
    integration.  The output is a three dimensional numpy array with 
    dimensions baselines, uv, and 512 points along the track ellipses.  
    Unlike computeUVW, however, only a single frequency (in Hz) can be 
    specified.
    
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.station` module instead of a list of stand ID
        numbers.
        
    .. versionchanged:: 1.0.0
        Added a keyword (site) to specify the station used for the 
        observation.
    """
    
    N = len(antennas)
    Nbase = N*(N-1)/2
    uvTrack = numpy.zeros((Nbase,2,512))
    
    # Phase center coordinates
    # Convert numbers to radians and, for HA, hours to degrees
    dec2 = dec * deg2rad
    lat2 = site.lat
    
    # Coordinate transformation matrices
    trans1 = numpy.matrix([[0, -numpy.sin(lat2), numpy.cos(lat2)],
                    [1,  0,               0],
                    [0,  numpy.cos(lat2), numpy.sin(lat2)]])
                    
    count = 0
    for i,j in getBaselines(antennas, Indicies=True):
        # Go from a east, north, up coordinate system to a celestial equation, 
        # east, north celestial pole system
        xyzPrime = antennas[i].stand - antennas[j].stand
        xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])
        xyz = numpy.ravel(xyz)
        
        uRange = numpy.linspace(-numpy.sqrt(xyz[0]**2 + xyz[1]**2), numpy.sqrt(xyz[0]**2 + xyz[1]**2), num=256)
        vRange1 = numpy.sqrt(xyz[0]**2 + xyz[1]**2 - uRange**2)*numpy.sin(dec2) + xyz[2]*numpy.cos(dec2)
        vRange2 = -numpy.sqrt(xyz[0]**2 + xyz[1]**2 - uRange**2)*numpy.sin(dec2) + xyz[2]*numpy.cos(dec2)
        
        uvTrack[count,0,0:256] = uRange * freq / numpy.array(speedOfLight)
        uvTrack[count,1,0:256] = vRange1 * freq / numpy.array(speedOfLight)
        uvTrack[count,0,256:512] = uRange[::-1] * freq / numpy.array(speedOfLight)
        uvTrack[count,1,256:512] = vRange2[::-1] * freq / numpy.array(speedOfLight)
        count = count + 1
        
    return uvTrack



# -*- coding: utf-8 -*-

"""Unit test for lsl.skymap module."""

import unittest

from lsl import skymap, astro


__revision__  = "$Rev: 2394 $"
__version__   = "0.2"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class skymap_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.skymap
    module."""
    
    def test_SkyMapGSM_init(self):
        """Test skymap.SkyMapGSM class constructor method.""" 
        
        skymap.SkyMapGSM()
        
    def test_SkyMapGSM_ComputeTotalPowerFromSky(self):
        """Test skymap.SkyMapGSM.ComputeTotalPowerFromSky() method."""
        
        s = skymap.SkyMapGSM()
        s.ComputeTotalPowerFromSky()
        
    def test_ProjectedSkyMap_init_GSM(self):
        """Test skymap.ProjectedSkyMap constructor method using SkyMapGSM."""
        
        s = skymap.SkyMapGSM()
        skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys())
        
    def test_ProjectedSkyMap_ComputeTotalPowerFromVisibleSky_GSM(self):
        """Test skymap.ProjectedSkyMap.ComputeTotalPowerFromVisibleSky() method using SkyMapGSM."""
        
        s = skymap.SkyMapGSM()
        p = skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys()) 
        p.ComputeTotalPowerFromVisibleSky()
        
    def test_ProjectedSkyMap_ComputeDirectionCosines_GSM(self):
        """Test skymap.ProjectedSkyMap.ComputeDirectionCosines() method using SkyMapGSM."""
        
        s = skymap.SkyMapGSM()
        p = skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys())
        p.ComputeDirectionCosines()
        
    def test_SkyMapLFSM_init(self):
        """Test skymap.SkyMapLFSM class constructor method.""" 
        
        skymap.SkyMapLFSM()
        
    def test_SkyMapLFSM_ComputeTotalPowerFromSky(self):
        """Test skymap.SkyMapLFSM.ComputeTotalPowerFromSky() method."""
        
        s = skymap.SkyMapLFSM()
        s.ComputeTotalPowerFromSky()

    def test_ProjectedSkyMap_init_LFSM(self):
        """Test skymap.ProjectedSkyMap constructor method using SkyMapLFSM."""
        
        s = skymap.SkyMapLFSM()
        skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys())

    def test_ProjectedSkyMap_ComputeTotalPowerFromVisibleSky_LFSM(self):
        """Test skymap.ProjectedSkyMap.ComputeTotalPowerFromVisibleSky() method using SkyMapLFSM."""
        
        s = skymap.SkyMapLFSM()
        p = skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys()) 
        p.ComputeTotalPowerFromVisibleSky()

    def test_ProjectedSkyMap_ComputeDirectionCosines_LFSM(self):
        """Test skymap.ProjectedSkyMap.ComputeDirectionCosines() method using SkyMapLFSM."""
        
        s = skymap.SkyMapLFSM()
        p = skymap.ProjectedSkyMap(s, 20, 30, astro.get_julian_from_sys())
        p.ComputeDirectionCosines()


class skymap_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lwa_user.skymap
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(skymap_tests))        
        
        
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.astro module."""


import unittest
import math
import pickle
import operator

from lsl import astro


__revision__  = "$Rev: 2470 $"
__version__   = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class astro_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.astro
    module."""

    def setUp(self):
        """
        Setup for running unit tests.
        """
        
        lng = astro.dms(True,  90, 0, 0)
        lat = astro.dms(False, 40, 0, 0)
        self.geo = astro.geo_posn(lng, lat)
        
        self.times = (\
            astro.date(2001,  1, 22, 0, 0, 0),
            astro.date(2001,  4, 16, 0, 0, 0),
            astro.date(2001,  8,  2, 0, 0, 0),
            astro.date(2001, 10, 10, 0, 0, 0),
            astro.date(1979,  2, 28, 0, 0, 0))
            
    def test_hms_init(self):
        """Test astro.hms constructor."""
        
        h = astro.hms(3, 42, 20.3242)
        self.assertEqual(h.hours, 3)
        self.assertEqual(h.minutes, 42)
        self.assertAlmostEqual(h.seconds, 20.3242)
            
        self.assertRaises(ValueError, astro.hms, 30, 0, 0)
        self.assertRaises(ValueError, astro.hms, 0, 90, 0)
        self.assertRaises(ValueError, astro.hms, 0, 0, 100) 
        
    def test_hms_reduce(self):
        """Test astro.hms.__reduce__() method."""
        
        h = astro.hms(3, 42, 20.3242)
        s = pickle.dumps(h)
        h = pickle.loads(s)
        self.assertEqual(h.hours, 3)   
        self.assertEqual(h.minutes, 42)
        self.assertAlmostEqual(h.seconds, 20.3242)
        
    def test_hms_cmp(self):
        """Test astro.hms.__cmp__() method."""
        
        h1 = astro.hms(2, 43, 32.232)
        h2 = astro.hms(2, 43, 32.232)
        h3 = astro.hms(4, 23, 42.221)
        
        self.assertTrue(operator.eq(h1, h2))
        self.assertFalse(operator.eq(h1, h3))
        self.assertFalse(operator.ne(h1, h2))
        self.assertTrue(operator.ne(h1, h3))
        self.assertTrue(operator.lt(h1, h3))
        self.assertTrue(operator.gt(h3, h1))
        self.assertTrue(operator.eq(h1, h1.to_deg()))
        self.assertTrue(operator.eq(h2, h2.to_dms()))
        
        self.assertRaises(TypeError, operator.eq, h3, "string")
        
    def test_hms_to_deg(self):
        """Test astro.hms_to_deg() function."""
        
        h = astro.hms(0, 0, 0)
        self.assertAlmostEqual(astro.hms_to_deg(h), 0.0)
        
        h = astro.hms(6, 0, 0)
        self.assertAlmostEqual(astro.hms_to_deg(h), 90.0)
        
        h = astro.hms(12, 0, 0)
        self.assertAlmostEqual(astro.hms_to_deg(h), 180.0)
        
        h = astro.hms(18, 0, 0)
        self.assertAlmostEqual(astro.hms_to_deg(h), 270.0)
        
    def test_deg_to_hms(self):
        """Test astro.deg_to_hms() function."""
        
        h = astro.deg_to_hms(0.0)
        self.assertEqual(h.hours, 0)
        self.assertEqual(h.minutes, 0)
        self.assertAlmostEqual(h.seconds, 0.0)
        
        h = astro.deg_to_hms(90.0)
        self.assertEqual(h.hours, 6)
        self.assertEqual(h.minutes, 0)
        self.assertAlmostEqual(h.seconds, 0.0)
        
        h= astro.deg_to_hms(180.0)
        self.assertEqual(h.hours, 12)
        self.assertEqual(h.minutes, 0)
        self.assertAlmostEqual(h.seconds, 0.0)
        
        h = astro.deg_to_hms(270.0)
        self.assertEqual(h.hours, 18)
        self.assertEqual(h.minutes, 0)
        self.assertAlmostEqual(h.seconds, 0.0)
        
    def test_dms_init(self):
        """Test astro.dms constructor."""
        
        d = astro.dms(True, 1, 30, 29.3245)
        self.assertTrue(d.neg)
        self.assertEqual(d.degrees, 1)
        self.assertEqual(d.minutes, 30)
        self.assertAlmostEqual(d.seconds, 29.3245)
        
        self.assertRaises(ValueError, astro.dms, False, 400, 0, 0)
        self.assertRaises(ValueError, astro.dms, False, 0, 80, 0)
        self.assertRaises(ValueError, astro.dms, False, 0, 0, 500)
        
    def test_dms_reduce(self):
        """Test astro.dms.__reduce__() method."""
        
        d = astro.dms(True, 1, 30, 29.3245)
        s = pickle.dumps(d)
        d = pickle.loads(s)
        self.assertTrue(d.neg)
        self.assertEqual(d.degrees, 1)
        self.assertEqual(d.minutes, 30)
        self.assertAlmostEqual(d.seconds, 29.3245)
        
    def test_dms_cmp(self):
        """Test astro.dms.__cmp__() method."""
        
        d1 = astro.dms(False, 2, 43, 32.232)
        d2 = astro.dms(False, 2, 43, 32.232)
        d3 = astro.dms(False, 4, 23, 42.221)
        
        self.assertTrue(operator.eq(d1, d2))
        self.assertFalse(operator.eq(d1, d3))
        self.assertFalse(operator.ne(d1, d2))
        self.assertTrue(operator.ne(d1, d3)) 
        self.assertTrue(operator.lt(d1, d3))
        self.assertTrue(operator.gt(d3, d1))
        self.assertTrue(operator.eq(d1, d1.to_deg())) 
        self.assertTrue(operator.eq(d2, d2.to_hms()))
        
        self.assertRaises(TypeError, operator.eq, d3, "string")
        
    def test_dms_to_deg(self):
        """Test astro.dms_to_deg() function."""
        
        d = astro.dms(False, 0, 0, 0)
        self.assertAlmostEqual(astro.dms_to_deg(d), 0.0)
        
        d = astro.dms(False, 100, 0, 0)
        self.assertAlmostEqual(astro.dms_to_deg(d), 100.0)
        
        d = astro.dms(True, 100, 0, 0)
        self.assertAlmostEqual(astro.dms_to_deg(d), -100.0)
        
    def test_deg_to_dms(self):
        """Test astro.deg_to_dms() function."""
        
        d = astro.deg_to_dms(0.0)
        self.assertEqual(d.neg, 0)
        self.assertEqual(d.degrees, 0)
        self.assertEqual(d.minutes, 0)
        self.assertAlmostEqual(d.seconds, 0.0)
        
        d = astro.deg_to_dms(100.0)
        self.assertEqual(d.neg, 0)
        self.assertEqual(d.degrees, 100)
        self.assertEqual(d.minutes, 0)
        self.assertAlmostEqual(d.seconds, 0.0)
        
        d = astro.deg_to_dms(-100.0)
        self.assertEqual(d.neg, 1)
        self.assertEqual(d.degrees, 100)
        self.assertEqual(d.minutes, 0)
        self.assertAlmostEqual(d.seconds, 0.0)
        
    def test_dir_cos(self):
        """Test astro.dir_cos() function."""
        
        h = astro.hrz_posn(0.0, 0.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, 0.0)
        self.assertAlmostEqual(m, 1.0)
        self.assertAlmostEqual(n, 0.0)
        
        h = astro.hrz_posn(90.0, 0.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, 1.0)
        self.assertAlmostEqual(m, 0.0)
        self.assertAlmostEqual(n, 0.0)
        
        h = astro.hrz_posn(180.0, 0.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, 0.0)
        self.assertAlmostEqual(m, -1.0)
        self.assertAlmostEqual(n, 0.0)
        
        h = astro.hrz_posn(270.0, 0.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, -1.0)
        self.assertAlmostEqual(m, 0.0)
        self.assertAlmostEqual(n, 0.0)
        
        h = astro.hrz_posn(0.0, 90.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, 0.0)
        self.assertAlmostEqual(m, 0.0)
        self.assertAlmostEqual(n, 1.0)
        
        h = astro.hrz_posn(0.0, -90.0)
        (l,m,n) = astro.dir_cos(h)
        self.assertAlmostEqual(l, 0.0)
        self.assertAlmostEqual(m, 0.0)
        self.assertAlmostEqual(n, -1.0)
        
    def test_range_degrees(self):
        """Test astro.range_degrees() function."""
        
        self.assertAlmostEqual(astro.range_degrees(20.0), 20.0)
        self.assertAlmostEqual(astro.range_degrees(370.0), 10.0)
        self.assertAlmostEqual(astro.range_degrees(-10.0), 350.0)
        
    def test_range_hours(self):
        """Test astro.range_hours() function."""
        
        self.assertAlmostEqual(astro.range_hours(10.0), 10.0)
        self.assertAlmostEqual(astro.range_hours(30.0), 6.0)
        self.assertAlmostEqual(astro.range_hours(-6.0), 18.0)
        
    def test_date_init(self):
        """Test astro.date constructor."""
        
        d = astro.date(2000, 4, 28, 21, 49, 13.0238)
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238)
        
        self.assertRaises(ValueError, astro.date, 2000, 30, 1, 0, 0, 0)
        self.assertRaises(ValueError, astro.date, 2000, 1, 48, 0, 0, 0)
        self.assertRaises(ValueError, astro.date, 2000, 1, 1, 39, 0, 0)
        self.assertRaises(ValueError, astro.date, 2000, 1, 1, 0, 69, 0)
        self.assertRaises(ValueError, astro.date, 2000, 1, 1, 0, 0, 73)
        
    def test_date_reduce(self):
        """Test astro.date.__reduce__() method."""
        
        d = astro.date(2000, 4, 28, 21, 49, 13.0238)
        s = pickle.dumps(d)
        d = pickle.loads(s)
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238)
        
    def test_date_cmp(self):
        """Test astro.date.__cmp__() method."""
        
        d1 = astro.date(2004, 5, 9, 12, 34, 22)
        d2 = astro.date(2009, 2, 6,  9,  2,  6)
        d3 = astro.date(2009, 2, 6,  9,  2,  6)
        self.assertTrue(operator.lt(d1, d2))
        self.assertTrue(operator.gt(d2, d1))
        self.assertTrue(operator.eq(d2, d3))
        self.assertTrue(operator.ne(d1, d3))
        
    def test_date_load(self):
        """Test astro.date.load() method."""
        
        d = astro.date()
        d.load('2000-4-28', '21:49:13.0238')
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238)
        
    def test_zonedate_init(self):
        """Test astro.zonedate constructor."""
        
        d = astro.zonedate(2000, 4, 28, 21, 49, 13.0238, -3600 * 5)
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238)
        self.assertEqual(d.gmtoff, -3600 * 5)
        
        self.assertRaises(ValueError, astro.zonedate, 2000, 30, 1, 0, 0, 0, 0)
        self.assertRaises(ValueError, astro.zonedate, 2000, 1, 48, 0, 0, 0, 0)
        self.assertRaises(ValueError, astro.zonedate, 2000, 1, 1, 39, 0, 0, 0)
        self.assertRaises(ValueError, astro.zonedate, 2000, 1, 1, 0, 69, 0, 0)
        self.assertRaises(ValueError, astro.zonedate, 2000, 1, 1, 0, 0, 73, 0)
        
    def test_zonedate_reduce(self):
        """Test astro.zonedate reduce method."""
        
        d = astro.zonedate(2000, 4, 28, 21, 49, 13.0238, -3600 * 5)
        s = pickle.dumps(d)
        d = pickle.loads(s)
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238)
        self.assertEqual(d.gmtoff, -3600 * 5)
        
    def test_date_to_zonedate(self):
        """Test astro.date_to_zonedate() function."""
        
        ACCURACY = 6
        
        d = astro.date(2000, 4, 28, 21, 49, 13.0238)
        z = astro.date_to_zonedate(d, -18000)
        self.assertEqual(z.years, 2000)
        self.assertEqual(z.months, 4)
        self.assertEqual(z.days, 28)
        self.assertEqual(z.hours, 16)
        self.assertEqual(z.minutes, 49)
        self.assertAlmostEqual(z.seconds, 13.0238, ACCURACY)
        self.assertEqual(z.gmtoff, -18000)
        
    def test_zonedate_to_date(self):
        """Test astro.zonedate_to_date() function."""
        
        ACCURACY = 4
        
        z = astro.zonedate(2000, 4, 28, 16, 49, 13.0238, -18000)
        d = astro.zonedate_to_date(z)
        self.assertEqual(d.years, 2000)
        self.assertEqual(d.months, 4)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.hours, 21)
        self.assertEqual(d.minutes, 49)
        self.assertAlmostEqual(d.seconds, 13.0238, ACCURACY)
        
    def test_get_date(self):
        """Test astro.get_date() function."""
        
        year_out = (2001, 2001, 2001, 2001, 1979)
        month_out = (1, 4, 8, 10, 2)
        day_out = (22, 16, 2, 10, 28)
        
        iyear = iter(year_out)
        imonth = iter(month_out)
        iday = iter(day_out)
        for t in self.times:
            j = t.to_jd()
            d = astro.get_date(j)
            self.assertEqual(d.years, next(iyear))
            self.assertEqual(d.months, next(imonth))
            self.assertEqual(d.days, next(iday))
            self.assertEqual(d.hours, 0)
            self.assertEqual(d.minutes, 0)
            self.assertAlmostEqual(d.seconds, 0.0)
            
    def test_utc_to_tai(self):
        """Test astro.utc_to_tai() function."""
        
        ACCURACY = 4
        
        jd_aa = (\
            astro.date(1972, 3, 1, 0, 0, 0).to_jd(),
            astro.date(1978, 6, 1, 0, 0, 0).to_jd(),
            astro.date(1983, 10, 1, 0, 0, 0).to_jd(),
            astro.date(1992, 2, 1, 0, 0, 0).to_jd())
            
        at_aa = (\
            astro.sec_to_jd(10.0), 
            astro.sec_to_jd(17.0), 
            astro.sec_to_jd(22.0),
            astro.sec_to_jd(27.0))
            
        iat = iter(at_aa)
        for t in jd_aa:
            tai = astro.utc_to_tai(t)
            self.assertAlmostEqual(tai - t, next(iat), ACCURACY) 
            
    def test_tai_to_utc(self):
        """Test astro.tai_to_utc() function."""
        
        ACCURACY = 4
        
        jd_aa = (\
            astro.date(1972, 3, 1, 0, 0, 0).to_jd(),
            astro.date(1978, 6, 1, 0, 0, 0).to_jd(),
            astro.date(1983, 10, 1, 0, 0, 0).to_jd(),
            astro.date(1992, 2, 1, 0, 0, 0).to_jd())
            
        at_aa = (\
            -astro.sec_to_jd(10.0), 
            -astro.sec_to_jd(17.0), 
            -astro.sec_to_jd(22.0),
            -astro.sec_to_jd(27.0))
            
        iat = iter(at_aa)
        for t in jd_aa:
            utc = astro.tai_to_utc(t)
            self.assertAlmostEqual(utc - t, next(iat), ACCURACY)            
            
    def test_get_geo_from_rect(self):
        """Test astro.get_geo_from_rect() function."""
        
        rect = astro.rect_posn(-1602196.60, -5042313.47, 3553971.51)
        geo = astro.get_geo_from_rect(rect)
        self.assertAlmostEqual(geo.lng, 252.3723299788652)
        self.assertAlmostEqual(geo.lat, 34.068899974626)
        self.assertAlmostEqual(geo.elv, 2126.9966418296099)
        
        rect = astro.rect_posn(1117051.64, -4848777.05, 3976930.95) 
        geo = astro.get_geo_from_rect(rect)
        self.assertAlmostEqual(geo.lng, 282.97333338968951)
        self.assertAlmostEqual(geo.lat, 38.821666658918112)
        self.assertAlmostEqual(geo.elv, 29.999061864800751)
        
    def test_get_rect_from_geo(self):
        """Test astro.get_rect_from_geo() function."""
        
        geo = astro.geo_posn(252.3723299788652, 34.068899974626, 2126.9966418296099)
        rect = astro.get_rect_from_geo(geo)
        self.assertAlmostEqual(rect.X, -1602196.60)
        self.assertAlmostEqual(rect.Y, -5042313.47)
        self.assertAlmostEqual(rect.Z, 3553971.51)
        
        geo = astro.geo_posn(282.97333338968951, 38.821666658918112, 29.999061864800751)
        rect = astro.get_rect_from_geo(geo)
        self.assertAlmostEqual(rect.X, 1117051.64)
        self.assertAlmostEqual(rect.Y, -4848777.05)
        self.assertAlmostEqual(rect.Z, 3976930.95)          
        
    def test_get_julian_day(self):
        """Test astro.get_julian_day() function and astro.date.to_jd() method."""
        
        jd_aa = (2451931.5, 2452015.5, 2452123.5, 2452192.5, 2443932.5)
        
        ijd = iter(jd_aa)
        for t in self.times:
            j = t.to_jd()
            self.assertAlmostEqual(j, next(ijd))
            
    def test_get_mean_sidereal_time(self):
        """Test astro.get_mean_sidereal_time() function."""
        
        ACCURACY = 4
        
        gmst_aa = (\
            astro.hms( 8,  5, 39.1981).to_sec() / 3600.0, 
            astro.hms(13, 36, 49.8490).to_sec() / 3600.0, 
            astro.hms(20, 42, 37.8288).to_sec() / 3600.0,
            astro.hms( 1, 14, 40.1492).to_sec() / 3600.0,
            astro.hms(10, 28, 52.7540).to_sec() / 3600.0)
            
        igmst = iter(gmst_aa)
        for t in self.times:
            j = t.to_jd()
            st = astro.get_mean_sidereal_time(j)
            self.assertAlmostEqual(st, next(igmst), ACCURACY)
            
    def test_get_apparent_sidereal_time(self):
        """Test astro.get_apparent_sidereal_time() function."""
        
        ACCURACY = 3
        
        gast_aa = (\
            astro.hms( 8,  5, 38.2531).to_sec() / 3600.0, 
            astro.hms(13, 36, 48.7661).to_sec() / 3600.0,
            astro.hms(20, 42, 36.8591).to_sec() / 3600.0,
            astro.hms( 1, 14, 39.0438).to_sec() / 3600.0,
            astro.hms(10, 28, 52.5980).to_sec() / 3600.0)
            
        igast = iter(gast_aa)
        for t in self.times:
            j = t.to_jd()
            st = astro.get_apparent_sidereal_time(j)
            self.assertAlmostEqual(st, next(igast), ACCURACY)
            
    def test_get_local_sidereal_time(self):
        """Test astro.get_local_sidereal_time() function."""
        
        ACCURACY = 3
        
        lng = astro.deg_to_sec(self.geo.lng)
        
        gast_aa = (\
            astro.range_hours((astro.hms( 8,  5, 38.2531).to_sec() + lng) / 3600.0), 
            astro.range_hours((astro.hms(13, 36, 48.7661).to_sec() + lng) / 3600.0),
            astro.range_hours((astro.hms(20, 42, 36.8591).to_sec() + lng) / 3600.0),
            astro.range_hours((astro.hms( 1, 14, 39.0438).to_sec() + lng) / 3600.0),
            astro.range_hours((astro.hms(10, 28, 52.5980).to_sec() + lng) / 3600.0))
            
        igast = iter(gast_aa)
        for t in self.times:
            j = t.to_jd()
            st = astro.get_local_sidereal_time(self.geo.lng, j)
            self.assertAlmostEqual(st, next(igast), ACCURACY)
            
    def test_equ_posn_init(self):
        """Test astro.equ_posn constructor."""
        
        e = astro.equ_posn(39.221, -24.543)
        self.assertAlmostEqual(e.ra, 39.221)
        self.assertAlmostEqual(e.dec, -24.543)
        
        ra = astro.deg_to_hms(39.221)
        dec = astro.deg_to_dms(-24.543)
        e = astro.equ_posn(ra, dec)
        self.assertAlmostEqual(e.ra, 39.221)
        self.assertAlmostEqual(e.dec, -24.543)
        
        self.assertRaises(ValueError, astro.equ_posn, 400, 0)
        self.assertRaises(ValueError, astro.equ_posn, -1, 0)
        self.assertRaises(ValueError, astro.equ_posn, 0, 100)
        self.assertRaises(ValueError, astro.equ_posn, 0, -100)
        
    def test_equ_posn_reduce(self):
        """Test astro.equ_posn.__reduce__() method."""
        
        e = astro.equ_posn(39.221, -24.543)
        s = pickle.dumps(e)
        e = pickle.loads(s)
        self.assertAlmostEqual(e.ra, 39.221)
        self.assertAlmostEqual(e.dec, -24.543)
        
    def test_equ_posn_cmp(self):
        """Test astro.equ_posn.__eq__() and astro.equ_posn.__ne__()
        methods."""
        
        e1 = astro.equ_posn(43, 12)
        e2 = astro.equ_posn(292, -21)
        e3 = astro.equ_posn(292, -21)
        self.assertTrue(operator.ne(e1, e2))
        self.assertFalse(operator.eq(e1, e2))
        self.assertTrue(operator.eq(e2, e3))
        self.assertFalse(operator.ne(e2, e3))
        
    def test_equ_posn_subscr(self):
        """Test astro.equ_posn.__getitem__() and astro.equ_posn.__setitem__()
        methods."""
        
        e = astro.equ_posn(39.221, -24.543)
        self.assertAlmostEqual(e[0], 39.221)
        self.assertAlmostEqual(e[1], -24.543)
        
        e[0] = 132.043
        e[1] = 29.392
        self.assertAlmostEqual(e.ra, 132.043)
        self.assertAlmostEqual(e.dec, 29.392)
        
        self.assertRaises(ValueError, operator.itemgetter(2), e)
        self.assertRaises(ValueError, operator.itemgetter(-1), e)
        
    def test_equ_posn_format(self):
        """Test astro.equ_posn.format() method."""
        
        e = astro.equ_posn(39.221, -24.543)
        (ra, dec) = e.format()
        self.assertAlmostEqual(ra.to_deg(), 39.221)
        self.assertAlmostEqual(dec.to_deg(), -24.543)
        
    def test_hrz_posn_init(self):
        """Test astro.hrz_posn constructor."""
        
        h = astro.hrz_posn(39.221, 46.301)
        self.assertAlmostEqual(h.az, 39.221)
        self.assertAlmostEqual(h.alt, 46.301)
        
        self.assertRaises(ValueError, astro.hrz_posn, 400, 0)
        self.assertRaises(ValueError, astro.hrz_posn, -1, 0)
        self.assertRaises(ValueError, astro.hrz_posn, 0, 100)
        self.assertRaises(ValueError, astro.hrz_posn, 0, -100)
        
    def test_hrz_posn_reduce(self):
        """Test astro.hrz_posn.__reduce__() method."""
        
        h = astro.hrz_posn(39.221, 46.301)
        s = pickle.dumps(h)
        h = pickle.loads(s)
        self.assertAlmostEqual(h.az, 39.221)
        self.assertAlmostEqual(h.alt, 46.301)
        
    def test_hrz_posn_cmp(self):
        """Test astro.hrz_posn.__eq__() and astro.hrz_posn.__ne__() 
        methods."""
        
        h1 = astro.hrz_posn(300, 21)
        h2 = astro.hrz_posn(33, 43)
        h3 = astro.hrz_posn(33, 43)
        self.assertTrue(operator.ne(h1, h2))
        self.assertFalse(operator.eq(h1, h2))
        self.assertTrue(operator.eq(h2, h3))
        self.assertFalse(operator.ne(h2, h3))
        
    def test_hrz_posn_subscr(self):
        """Test astro.hrz_posn.__getitem__() and astro.hrz_posn.__setitem__() 
        methods."""
        
        h = astro.hrz_posn(39.221, 46.301)
        self.assertAlmostEqual(h[0], 39.221)
        self.assertAlmostEqual(h[1], 46.301)
        h[0] = 319.293
        h[1] = -42.212
        self.assertAlmostEqual(h.az, 319.293)
        self.assertAlmostEqual(h.alt, -42.212)
        
        self.assertRaises(ValueError, operator.itemgetter(2), h)
        self.assertRaises(ValueError, operator.itemgetter(-1), h)
        
    def test_gal_posn_init(self):
        """Test astro.gal_posn constructor."""
        
        g = astro.gal_posn(39.221, -24.543)
        self.assertAlmostEqual(g.l, 39.221)
        self.assertAlmostEqual(g.b, -24.543)
        
        l = astro.deg_to_dms(39.221)
        b = astro.deg_to_dms(-24.543)
        g = astro.gal_posn(l, b)
        self.assertAlmostEqual(g.l, 39.221)
        self.assertAlmostEqual(g.b, -24.543)
        
        self.assertRaises(ValueError, astro.gal_posn, 400, 0)
        self.assertRaises(ValueError, astro.gal_posn, -400, 0)
        self.assertRaises(ValueError, astro.gal_posn, 0, 100)
        self.assertRaises(ValueError, astro.gal_posn, 0, -100)
        
    def test_gal_posn_reduce(self):
        """Test astro.gal_posn.__reduce__() method."""
        
        g = astro.gal_posn(39.221, -24.543)
        s = pickle.dumps(g)
        g = pickle.loads(s)
        self.assertAlmostEqual(g.l, 39.221)
        self.assertAlmostEqual(g.b, -24.543)
        
    def test_gal_posn_cmp(self):
        """Test astro.gal_posn.__eq__() and astro.gal_posn.__ne__() 
        methods."""
        
        g1 = astro.gal_posn(232, 54)
        g2 = astro.gal_posn(123, -32)
        g3 = astro.gal_posn(123, -32)
        self.assertTrue(operator.ne(g1, g2))
        self.assertFalse(operator.eq(g1, g2))
        self.assertTrue(operator.eq(g2, g3))
        self.assertFalse(operator.ne(g2, g3))
        
    def test_gal_posn_subscr(self):
        """Test astro.gal_posn.__getitem__() and astro.gal_posn.__setitem__()
        methods."""
        
        g = astro.gal_posn(39.221, -24.543)
        self.assertAlmostEqual(g[0], 39.221)
        self.assertAlmostEqual(g[1], -24.543)
        
        g[0] = 125.345
        g[1] = 43.232
        self.assertAlmostEqual(g.l, 125.345)
        self.assertAlmostEqual(g.b, 43.232)
        
        self.assertRaises(ValueError, operator.itemgetter(3), g)
        self.assertRaises(ValueError, operator.itemgetter(-1), g)
        
    def test_gal_posn_format(self):
        """Test astro.gal_posn.format() method."""
        
        g = astro.gal_posn(39.221, -24.543)
        (l, b) = g.format()
        self.assertAlmostEqual(l.to_deg(), 39.221)
        self.assertAlmostEqual(b.to_deg(), -24.543)
        
    def test_geo_posn_init(self):
        """Test astro.geo_posn constructor."""
        
        g = astro.geo_posn(39.221, -24.543, 2000.345)
        self.assertAlmostEqual(g.lng, 39.221)
        self.assertAlmostEqual(g.lat, -24.543)
        self.assertAlmostEqual(g.elv, 2000.345)
        
        lng = astro.deg_to_dms(39.221)
        lat = astro.deg_to_dms(-24.543)
        g = astro.geo_posn(lng, lat, 2000.345)
        self.assertAlmostEqual(g.lng, 39.221)
        self.assertAlmostEqual(g.lat, -24.543)
        self.assertAlmostEqual(g.elv, 2000.345)
        
        self.assertRaises(ValueError, astro.geo_posn, 400, 0)
        self.assertRaises(ValueError, astro.geo_posn, -400, 0)
        self.assertRaises(ValueError, astro.geo_posn, 0, 100)
        self.assertRaises(ValueError, astro.geo_posn, 0, -100)
        
    def test_geo_posn_reduce(self):
        """Test astro.geo_posn.__reduce__() method."""
        
        g = astro.geo_posn(39.221, -24.543, 2000.345)
        s = pickle.dumps(g)
        g = pickle.loads(s)
        self.assertAlmostEqual(g.lng, 39.221)
        self.assertAlmostEqual(g.lat, -24.543)
        self.assertAlmostEqual(g.elv, 2000.345)
        
    def test_geo_posn_cmp(self):
        """Test astro.geo_posn.__eq__() and astro.geo_posn.__ne__()
        methods."""
        
        g1 = astro.geo_posn(232, 53, 1322)
        g2 = astro.geo_posn(124, -24, 1243)
        g3 = astro.geo_posn(124, -24, 1243)
        self.assertTrue(operator.ne(g1, g2))
        self.assertFalse(operator.eq(g1, g2))
        self.assertTrue(operator.eq(g2, g3))
        self.assertFalse(operator.ne(g2, g3))
        
    def test_geo_posn_subscr(self):
        """Test astro.geo_posn.__getitem__() and astro.geo_posn.__setitem__()
        methods."""
        
        g = astro.geo_posn(39.221, -24.543, 2000.345)
        self.assertAlmostEqual(g[0], 39.221)
        self.assertAlmostEqual(g[1], -24.543)
        self.assertAlmostEqual(g[2], 2000.345)
        
        g[0] = 200.221
        g[1] = 38.231
        g[2] = 1003.053
        self.assertAlmostEqual(g.lng, 200.221)
        self.assertAlmostEqual(g.lat, 38.231)
        self.assertAlmostEqual(g.elv, 1003.053)
        
        self.assertRaises(ValueError, operator.itemgetter(3), g)
        self.assertRaises(ValueError, operator.itemgetter(-1), g)
        
    def test_geo_posn_format(self):
        """Test astro.geo_posn.format() method."""
        
        g = astro.geo_posn(39.221, -24.543, 2000.345)
        (lng, lat) = g.format()
        self.assertAlmostEqual(lng.to_deg(), 39.221)
        self.assertAlmostEqual(lat.to_deg(), -24.543)
        
    def test_rect_posn_init(self):
        """Test astro.rect_posn constructor method."""
        
        r = astro.rect_posn(2.409, 9.324, 4.442)
        self.assertAlmostEqual(r.X, 2.409)
        self.assertAlmostEqual(r.Y, 9.324)
        self.assertAlmostEqual(r.Z, 4.442)
        
    def test_rect_posn_reduce(self):
        """Test astro.rect_posn pickle.__reduce__() method."""
        
        r = astro.rect_posn(2.409, 9.324, 4.442)
        s = pickle.dumps(r)
        r = pickle.loads(s)
        self.assertAlmostEqual(r.X, 2.409)
        self.assertAlmostEqual(r.Y, 9.324)
        self.assertAlmostEqual(r.Z, 4.442)
        
    def test_rect_posn_cmp(self):
        """Test astro.rect_posn.__eq__() and astro.rect_posn.__ne__()
        methods."""
        
        r1 = astro.rect_posn(4, 5, 2)
        r2 = astro.rect_posn(5, 3, 1)
        r3 = astro.rect_posn(5, 3, 1)
        self.assertTrue(operator.ne(r1, r2))
        self.assertFalse(operator.eq(r1, r2))
        self.assertTrue(operator.eq(r2, r3))
        self.assertFalse(operator.ne(r2, r3))
        
    def test_rect_posn_subscr(self):
        """Test astro.rect_posn.__getitem__() and astro.rect_posn.__setitem__()
        methods."""
        
        r = astro.rect_posn(2.409, 9.324, 4.442)
        self.assertAlmostEqual(r[0], 2.409)
        self.assertAlmostEqual(r[1], 9.324)
        self.assertAlmostEqual(r[2], 4.442)
        r[0] = -2.343
        r[1] = -4.215
        r[2] = 0.432
        self.assertAlmostEqual(r.X, -2.343)
        self.assertAlmostEqual(r.Y, -4.215)
        self.assertAlmostEqual(r.Z, 0.432)
        
    def test_get_hrz_from_equ(self):
        """Test astro.get_hrz_from_equ() function."""
        
        ACCURACY_AZ = 0
        ACCURACY_ALT = 0
        
        ra_aa = (\
            astro.hms( 8,  5, 38.2531).to_deg(), 
            astro.hms(13, 36, 48.7661).to_deg(),
            astro.hms(20, 42, 36.8591).to_deg(),
            astro.hms( 1, 14, 39.0438).to_deg(),
            astro.hms(10, 28, 52.5980).to_deg())
            
        obs = astro.lnlat_posn(0.0, 0.0)    
        
        ira = iter(ra_aa)
        for t in self.times:
            ra = next(ira)
            
            j = t.to_jd()
            equ = astro.equ_posn(ra, 30.0)
            hrz = astro.get_hrz_from_equ(equ, obs, j)
            az = hrz.az
            if az > 180:
                    az -= 360
            self.assertAlmostEqual(math.radians(az), 0.0, ACCURACY_AZ)
            self.assertAlmostEqual(hrz.alt, 60.0, ACCURACY_ALT)               
            
    def test_get_equ_from_hrz(self):
        """Test astro.get_equ_from_hrz() function."""
        
        ACCURACY_RA = 0
        ACCURACY_DEC = 0
        
        ra_aa = (\
            astro.hms( 8,  5, 38.2531).to_deg(), 
            astro.hms(13, 36, 48.7661).to_deg(),
            astro.hms(20, 42, 36.8591).to_deg(),
            astro.hms( 1, 14, 39.0438).to_deg(),
            astro.hms(10, 28, 52.5980).to_deg()) 
            
        hrz = astro.hrz_posn(0.0, 60.0)
        obs = astro.lnlat_posn(0.0, 0.0)
        
        ira = iter(ra_aa)
        for t in self.times:
            ra = next(ira)
            
            j = t.to_jd()
            equ = astro.get_equ_from_hrz(hrz, obs, j)
            self.assertAlmostEqual(math.radians(equ.ra), math.radians(ra), ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, 30.0, ACCURACY_DEC)
            
    def test_get_equ_prec(self):
        """Test astro.get_equ_prec() function."""
        
        ACCURACY_RA = 4
        ACCURACY_DEC = 4
        
        equ_in = (\
            astro.equ_posn(  0.0,   0.0),
            astro.equ_posn(180.0,   0.0), 
            astro.equ_posn(  0.0,  80.0),
            astro.equ_posn(240.0, -55.0),
            astro.equ_posn(  0.0,   0.0))
            
        iequ = iter(equ_in)
        for t in self.times:
            j = t.to_jd() 
            equ0 = next(iequ)
            equ = astro.get_equ_prec(equ0, j)
            chk = astro.get_precession(astro.J2000_UTC_JD, equ0, j)
            self.assertAlmostEqual(equ.ra, chk.ra, ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, chk.dec, ACCURACY_DEC) 
            
    def test_get_equ_prec2(self):
        """Test astro.get_equ_prec2() function."""
        
        ACCURACY_RA = 4
        ACCURACY_DEC = 4
        
        equ_in = (\
            astro.equ_posn(  0.0,   0.0),
            astro.equ_posn(180.0,   0.0), 
            astro.equ_posn(  0.0,  80.0),
            astro.equ_posn(240.0, -55.0),
            astro.equ_posn(  0.0,   0.0))
            
        iequ = iter(equ_in)
        for t in self.times:
            j = t.to_jd() 
            equ0 = next(iequ)
            equ = astro.get_equ_prec2(equ0, j, astro.J2000_UTC_JD)
            chk = astro.get_precession(j, equ0, astro.J2000_UTC_JD)
            self.assertAlmostEqual(equ.ra, chk.ra, ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, chk.dec, ACCURACY_DEC)
            
    def test_nutation_init(self):
        """Test astro.nutation constructor."""
        
        n = astro.nutation(0.0234, -0.0421, 23.5656)
        self.assertAlmostEqual(n.longitude, 0.0234)
        self.assertAlmostEqual(n.obliquity, -0.0421)
        self.assertAlmostEqual(n.ecliptic, 23.5656)
        
        self.assertRaises(ValueError, astro.nutation, 400, 0, 0)
        self.assertRaises(ValueError, astro.nutation, -400, 0, 0)
        self.assertRaises(ValueError, astro.nutation, 0, 100, 0)
        self.assertRaises(ValueError, astro.nutation, 0, -100, 0)
        self.assertRaises(ValueError, astro.nutation, 0, 0, 100)
        self.assertRaises(ValueError, astro.nutation, 0, 0, -100)
        
    def test_get_nutation(self):
        """Test astro.get_nutation() function."""
        
        ACCURACY_LNG = 4
        ACCURACY_OBL = 4
        ACCURACY_ECL = 4
        
        lng_aa = (\
            astro.dms(True,  0, 0, 15.453).to_deg(),
            astro.dms(True,  0, 0, 17.708).to_deg(),
            astro.dms(True,  0, 0, 15.856).to_deg(),
            astro.dms(True,  0, 0, 18.075).to_deg(),
            astro.dms(True,  0, 0,  2.557).to_deg())
            
        obl_aa = (\
            astro.dms(True,  0, 0, 2.569).to_deg(),
            astro.dms(True,  0, 0, 1.241).to_deg(),
            astro.dms(True,  0, 0, 0.797).to_deg(),
            astro.dms(False, 0, 0, 0.339).to_deg(),
            astro.dms(True,  0, 0, 8.606).to_deg())
            
        ecl_aa = (\
            astro.dms(False, 23, 26, 18.384).to_deg(),
            astro.dms(False, 23, 26, 19.604).to_deg(),
            astro.dms(False, 23, 26, 19.909).to_deg(),
            astro.dms(False, 23, 26, 20.957).to_deg(),
            astro.dms(False, 23, 26, 22.569).to_deg())
            
        ilng = iter(lng_aa)
        iobl = iter(obl_aa)
        iecl = iter(ecl_aa)
        for t in self.times:
            lng = next(ilng)
            obl = next(iobl)
            ecl = next(iecl)
            
            j = t.to_jd()
            nut = astro.get_nutation(j)
            self.assertAlmostEqual(nut.longitude, lng, ACCURACY_LNG)      
            self.assertAlmostEqual(nut.obliquity, obl, ACCURACY_OBL)
            self.assertAlmostEqual(nut.ecliptic, ecl, ACCURACY_ECL)
            
    def test_get_equ_nut(self):
        """Test astro.get_equ_nut() function."""
        
        body = astro.equ_posn(41.547213, 49.348483)
        JD = 2462088.7
        
        ACCURACY = 3
        deltaRA = 15.842766509979356
        deltaDec = 6.2164102884679551
        
        body2 = astro.get_equ_nut(body, JD)
        self.assertAlmostEqual((body2.ra -body.ra )*3600, deltaRA,  ACCURACY)
        self.assertAlmostEqual((body2.dec-body.dec)*3600, deltaDec, ACCURACY)
        
    def get_equ_aber(self):
        """Test astro.get_equ_aber() funciton."""
        
        body = astro.equ_posn(41.547213, 49.348483)
        JD = 2462088.7
        
        ACCURACY = 3
        deltaRA = 30.09128661282716
        deltaDec = 6.621256621590987
        
        body2 = astro.get_equ_aber(body, JD)
        self.assertAlmostEqual((body2.ra -body.ra )*3600, deltaRA,  ACCURACY)
        self.assertAlmostEqual((body2.dec-body.dec)*3600, deltaDec, ACCURACY)
        
    def test_rst_time_init(self):
        """Test astro.rst_time constructor method."""
        
        rise = self.times[0].to_jd()
        transit = self.times[1].to_jd()
        set = self.times[2].to_jd()
        rst = astro.rst_time(rise, set, transit)
        self.assertAlmostEqual(rst.rise, rise)
        self.assertAlmostEqual(rst.transit, transit)
        self.assertAlmostEqual(rst.set, set)
        
    def test_get_object_rst(self):
        """Test astro.get_object_rst() function."""
        
        body = astro.equ_posn(0, 33.0)
        
        ACCURACY = 2
        transit_aa = (\
            astro.date(2001,  1, 22, 21, 50, 48.04).to_jd(),
            astro.date(2001,  4, 16, 16, 20, 31.74).to_jd(),
            astro.date(2001,  8,  2,  9, 15, 56.73).to_jd(),
            astro.date(2001, 10, 10,  4, 44, 39.94).to_jd(),
            astro.date(1979,  2, 28, 19, 26, 50.45).to_jd())
            
        itr = iter(transit_aa) 
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_object_rst(j, self.geo, body)
            self.assertAlmostEqual(astro.utc_to_tt(trns.transit), next(itr), ACCURACY)
            
        body = astro.equ_posn(0, 89.0)
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_object_rst(j, self.geo, body)
            self.assertTrue(trns is None)
            
        body = astro.equ_posn(0, -89.0)
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_object_rst(j, self.geo, body)
            self.assertTrue(trns is None)
            
    def test_get_solar_rst(self):
        """Test astro.get_solar_rst() function."""
        
        ACCURACY = 2
        
        p = (-self.geo.lng / 360.0) + (1.002738 * astro.sec_to_jd(64)) 
        transit_aa = (\
            astro.date(2001,  1, 22, 12, 11, 37.57).to_jd() + p,
            astro.date(2001,  4, 16, 11, 59, 44.68).to_jd() + p,
            astro.date(2001,  8,  2, 12,  6, 14.09).to_jd() + p,
            astro.date(2001, 10, 10, 11, 46, 58.27).to_jd() + p,
            astro.date(1979,  2, 28, 12, 12, 29.32).to_jd() + p)
            
        itr = iter(transit_aa) 
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_solar_rst(j, self.geo)
            self.assertAlmostEqual(astro.utc_to_tt(trns.transit), next(itr), ACCURACY)
            
    def test_get_solar_equ_coords(self):
        """Test astro.get_solar_equ_coords() function."""
        
        ACCURACY_RA  = 2
        ACCURACY_DEC = 2
        
        ra_aa = (\
            astro.hms(20, 17,  7.63).to_deg(),
            astro.hms( 1, 36, 40.54).to_deg(),
            astro.hms( 8, 48, 53.15).to_deg(),
            astro.hms(13,  1, 45.13).to_deg(),
            astro.hms(22, 41, 38.54).to_deg())
            
        dec_aa = (\
            astro.dms(True,  19, 42, 30.7).to_deg(),
            astro.dms(False, 10,  3, 55.1).to_deg(),
            astro.dms(False, 17, 47, 59.1).to_deg(),
            astro.dms(True,   6, 34, 59.9).to_deg(),
            astro.dms(True,   8, 16, 14.4).to_deg())     
            
        ira = iter(ra_aa)
        idec = iter(dec_aa)
        for t in self.times:
            j = t.to_jd()
            equ = astro.get_solar_equ_coords(j)
            self.assertAlmostEqual(equ.ra, next(ira), ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, next(idec), ACCURACY_DEC)
            
    def test_get_jupiter_rst(self):
        """Test astro.get_jupiter_rst() function."""
        
        ACCURACY = 2
        
        p = (-self.geo.lng / 360.0) + (1.002738 * astro.sec_to_jd(64))
        transit_aa = (\
            astro.date(2001,  1, 21, 19, 51, 59).to_jd() + p,
            astro.date(2001,  4, 16, 14, 56, 44).to_jd() + p,
            astro.date(2001,  8,  2,  9, 34, 53).to_jd() + p,
            astro.date(2001, 10, 10,  5, 48, 36).to_jd() + p,
            astro.date(1979,  2, 27, 21, 36, 57).to_jd() + p)
            
        itr = iter(transit_aa)
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_jupiter_rst(j, self.geo)
            self.assertAlmostEqual(astro.utc_to_tt(trns.transit), next(itr), ACCURACY)
            
    def test_get_jupiter_equ_coord(self):
        """Test astro.get_jupiter_equ_coords() function."""
        
        ACCURACY_RA  = 2
        ACCURACY_DEC = 2
        
        ra_aa = (\
            astro.hms( 3, 56, 56.236).to_deg(),
            astro.hms( 4, 35, 28.272).to_deg(),
            astro.hms( 6, 18, 42.909).to_deg(),
            astro.hms( 7,  4,  7.660).to_deg(),
            astro.hms( 8,  9, 40.552).to_deg()) 
            
        dec_aa = (\
            astro.dms(False, 19, 41, 39.73).to_deg(),
            astro.dms(False, 21, 34, 42.35).to_deg(),
            astro.dms(False, 23,  7,  6.00).to_deg(),
            astro.dms(False, 22, 27, 30.44).to_deg(),
            astro.dms(False, 20, 48, 29.55).to_deg())  
            
        ira = iter(ra_aa)
        idec = iter(dec_aa)
        for t in self.times:
            j = t.to_jd()
            equ = astro.get_jupiter_equ_coords(j)
            self.assertAlmostEqual(equ.ra, next(ira), ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, next(idec), ACCURACY_DEC)                                                     
            
    def test_get_lunar_rst(self):
        """Test astro.get_lunar_rst() function."""
        
        ACCURACY = 1
        
        p = (-self.geo.lng / 360.0) + (1.002738 * astro.sec_to_jd(64))
        transit_aa = (\
            astro.date(2001,  1, 22, 10, 34, 47).to_jd() + p,
            astro.date(2001,  4, 16,  6, 46, 16).to_jd() + p,
            astro.date(2001,  8,  1, 22, 19, 42).to_jd() + p,
            astro.date(2001, 10, 10,  6,  3, 51).to_jd() + p,
            astro.date(1979,  2, 28, 13, 51, 32).to_jd() + p)
            
        itr = iter(transit_aa)
        for t in self.times:
            j = t.to_jd()
            trns = astro.get_lunar_rst(j, self.geo)
            self.assertAlmostEqual(astro.utc_to_tt(trns.transit), next(itr), ACCURACY)
            
    def test_get_lunar_equ_coords(self):
        """Test astro.get_lunar_equ_coords() function."""
        
        ACCURACY_RA  = 1
        ACCURACY_DEC = 2
        
        ra_aa = (\
            astro.hms(18, 19, 24.98).to_deg(),
            astro.hms(20,  9, 55.87).to_deg(),
            astro.hms(19,  5, 43.33).to_deg(),
            astro.hms( 7,  3, 20.27).to_deg(),
            astro.hms(23, 50, 17.96).to_deg())
            
        dec_aa = (\
            astro.dms(True,  22, 20, 47.9).to_deg(),
            astro.dms(True,  21, 54, 24.7).to_deg(),
            astro.dms(True,  23, 23, 33.8).to_deg(),
            astro.dms(False, 23, 54,  8.3).to_deg(),
            astro.dms(True,   2,  0, 28.9).to_deg())    
            
        ira = iter(ra_aa)
        idec = iter(dec_aa)
        for t in self.times:
            j = t.to_jd()
            equ = astro.get_lunar_equ_coords(j)
            self.assertAlmostEqual(equ.ra, next(ira), ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, next(idec), ACCURACY_DEC) 
            
    def test_get_ecl_from_equ(self):
        """Test astro.get_ecl_from_equ() function."""
        
        ACCURACY_LNG = 4
        ACCURACY_LAT = 4
        
        # Data set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        lng_aa = (302.1000, 26.3000, 129.5000, 196.7000, 338.9000)
        lat_aa = (  0.0000, -1.0000,   3.0000,  -5.0000,   7.0000)
        
        ra_aa = (\
            astro.hms(20, 17, 26.98).to_deg(),
            astro.hms( 1, 39,  1.88).to_deg(),
            astro.hms( 8, 51, 10.38).to_deg(),
            astro.hms(12, 53, 44.69).to_deg(),
            astro.hms(22, 31, 32.27).to_deg())
        dec_aa = (\
            astro.dms(True,  19, 41, 50.94).to_deg(),
            astro.dms(False,  9, 13, 17.35).to_deg(),
            astro.dms(False, 20, 46, 10.16).to_deg(),
            astro.dms(True,  11, 10, 41.32).to_deg(),
            astro.dms(True,   1, 44, 23.34).to_deg())  
            
        ilng = iter(lng_aa)
        ilat = iter(lat_aa)
        ira = iter(ra_aa)
        idec = iter(dec_aa)
        for t in self.times:
            lng = next(ilng)
            lat = next(ilat)
            ra = next(ira)
            dec = next(idec)
            
            j = t.to_jd()
            equ = astro.equ_posn(ra, dec)
            ecl = astro.get_ecl_from_equ(equ, j)
            self.assertAlmostEqual(ecl.lng, lng, ACCURACY_LNG)
            self.assertAlmostEqual(ecl.lat, lat, ACCURACY_LAT)
            
    def test_get_equ_from_ecl(self):
        """Test astro.get_equ_from_ecl() function."""
        
        ACCURACY_RA = 4
        ACCURACY_DEC = 4
        
        # Data set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        lng_aa = (302.1000, 26.3000, 129.5000, 196.7000, 338.9000)
        lat_aa = (  0.0000, -1.0000,   3.0000,  -5.0000,   7.0000)
        
        ra_aa = (\
            astro.hms(20, 17, 26.98).to_deg(),
            astro.hms( 1, 39,  1.88).to_deg(),
            astro.hms( 8, 51, 10.38).to_deg(),
            astro.hms(12, 53, 44.69).to_deg(),
            astro.hms(22, 31, 32.27).to_deg())
        dec_aa = (\
            astro.dms(True,  19, 41, 50.94).to_deg(),
            astro.dms(False,  9, 13, 17.35).to_deg(),
            astro.dms(False, 20, 46, 10.16).to_deg(),
            astro.dms(True,  11, 10, 41.32).to_deg(),
            astro.dms(True,   1, 44, 23.34).to_deg()) 
            
        ilng = iter(lng_aa)
        ilat = iter(lat_aa)
        ira = iter(ra_aa)
        idec = iter(dec_aa)
        for t in self.times:
            lng = next(ilng)
            lat = next(ilat)
            ra = next(ira)
            dec = next(idec)
            
            j = t.to_jd()
            
            ecl = astro.ecl_posn(lng, lat)
            equ = astro.get_equ_from_ecl(ecl, j)
            self.assertAlmostEqual(equ.ra, ra, ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, dec, ACCURACY_DEC)
            
    def test_get_gal_from_equ(self):
        """Test astro.get_gal_from_equ() function."""
        
        ACCURACY_L = 2
        ACCURACY_B = 2
        
        # Test set generated at http://fuse.pha.jhu.edu/support/tools/eqtogal.html
        ra_4c = (\
            astro.hms(13, 35, 31.41).to_deg(),
            astro.hms(15, 56,  3.01).to_deg(),
            astro.hms( 1, 35, 19.19).to_deg(),
            astro.hms(19, 45, 35.25).to_deg())  
        dec_4c = (\
            astro.dms(True,   6,  3, 38.0).to_deg(),
            astro.dms(False,  1, 49, 36.9).to_deg(),
            astro.dms(False, 22, 46,  3.9).to_deg(),
            astro.dms(False, 78, 37,  8.5).to_deg())
            
        l_4c = (323.3000, 11.5000, 136.7000, 110.9000)  
        b_4c = ( 54.7000, 38.5000, -38.6000,  24.1000)
        
        ira = iter(ra_4c)
        idec = iter(dec_4c)
        il = iter(l_4c)
        ib = iter(b_4c)
        for i in range(len(ra_4c)):
            ra = next(ira)
            dec = next(idec)
            l = next(il)
            b = next(ib)
            
            equ = astro.equ_posn(ra, dec)
            gal = astro.get_gal_from_equ(equ)
            self.assertAlmostEqual(gal.l, l, ACCURACY_L)
            self.assertAlmostEqual(gal.b, b, ACCURACY_B)
            
    def test_get_equ_from_gal(self):
        """Test astro.get_equ_from_gal() function."""
        
        ACCURACY_RA = 2
        ACCURACY_DEC = 2
        
        # Test set generated at http://fuse.pha.jhu.edu/support/tools/eqtogal.html
        ra_4c = (\
            astro.hms(13, 35, 31.41).to_deg(),
            astro.hms(15, 56,  3.01).to_deg(),
            astro.hms( 1, 35, 19.19).to_deg(),
            astro.hms(19, 45, 35.25).to_deg())  
        dec_4c = (\
            astro.dms(True,   6,  3, 38.0).to_deg(),
            astro.dms(False,  1, 49, 36.9).to_deg(),
            astro.dms(False, 22, 46,  3.9).to_deg(),
            astro.dms(False, 78, 37,  8.5).to_deg())
            
        l_4c = (323.3000, 11.5000, 136.7000, 110.9000)  
        b_4c = ( 54.7000, 38.5000, -38.6000,  24.1000)
        
        ira = iter(ra_4c)
        idec = iter(dec_4c)
        il = iter(l_4c)
        ib = iter(b_4c)
        for i in range(len(ra_4c)):
            ra = next(ira)
            dec = next(idec)
            l = next(il)
            b = next(ib)
            
            gal = astro.gal_posn(l, b)
            equ = astro.get_equ_from_gal(gal)
            self.assertAlmostEqual(equ.ra, ra, ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, dec, ACCURACY_DEC)
            
    def test_get_gal_from_equ2000(self):
        """Test astro.get_gal_from_equ2000() function."""
        
        ACCURACY_L = 2
        ACCURACY_B = 2
        
        # Test set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        ra_4c = (\
            astro.hms(13, 35, 31.41).to_deg(),
            astro.hms(15, 56,  3.01).to_deg(),
            astro.hms( 1, 35, 19.19).to_deg(),
            astro.hms(19, 45, 35.25).to_deg())  
        dec_4c = (\
            astro.dms(True,   6,  3, 38.0).to_deg(),
            astro.dms(False,  1, 49, 36.9).to_deg(),
            astro.dms(False, 22, 46,  3.9).to_deg(),
            astro.dms(False, 78, 37,  8.5).to_deg())
            
        l_4c = (322.3585, 11.1982, 135.9796, 110.8142)  
        b_4c = ( 55.1396, 39.1055, -38.9896,  23.9482)
        
        ira = iter(ra_4c)
        idec = iter(dec_4c)
        il = iter(l_4c)
        ib = iter(b_4c)
        for i in range(len(ra_4c)):
            ra = next(ira)
            dec = next(idec)
            l = next(il)
            b = next(ib)
            
            equ = astro.equ_posn(ra, dec)
            gal = astro.get_gal_from_equ2000(equ)
            self.assertAlmostEqual(gal.l, l, ACCURACY_L)
            self.assertAlmostEqual(gal.b, b, ACCURACY_B)
            
    def test_get_equ2000_from_gal(self):
        """Test astro.get_equ2000_from_gal() function."""
        
        ACCURACY_RA = 2
        ACCURACY_DEC = 2
        
        # Test set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        ra_4c = (\
            astro.hms(13, 35, 31.41).to_deg(),
            astro.hms(15, 56,  3.01).to_deg(),
            astro.hms( 1, 35, 19.19).to_deg(),
            astro.hms(19, 45, 35.25).to_deg())  
        dec_4c = (\
            astro.dms(True,   6,  3, 38.0).to_deg(),
            astro.dms(False,  1, 49, 36.9).to_deg(),
            astro.dms(False, 22, 46,  3.9).to_deg(),
            astro.dms(False, 78, 37,  8.5).to_deg())
            
        l_4c = (322.3585, 11.1982, 135.9796, 110.8142)  
        b_4c = ( 55.1396, 39.1055, -38.9896,  23.9482)
        
        ira = iter(ra_4c)
        idec = iter(dec_4c)
        il = iter(l_4c)
        ib = iter(b_4c)
        for i in range(len(ra_4c)):
            ra = next(ira)
            dec = next(idec)
            l = next(il)
            b = next(ib)
            
            gal = astro.gal_posn(l, b)
            equ = astro.get_equ2000_from_gal(gal)
            self.assertAlmostEqual(equ.ra, ra, ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, dec, ACCURACY_DEC)
            
    def test_get_apparent_posn(self):
        """Test astro.get_apparent_posn() function."""
        
        ACCURACY_RA = 4
        ACCURACY_DEC = 4
        
        ra_in  = (201.2500, 69.2517)  
        dec_in = (-43.0667,   29.6708)
        
        ra_out  = (201.26667, 69.27255)
        dec_out = (-43.06925, 29.67399)
        
        idec_in = iter(dec_in)
        ira_out = iter(ra_out)
        idec_out = iter(dec_out)
        j = self.times[0].to_jd()
        for ra in ra_in:
            equ = astro.equ_posn(ra, next(idec_in))
            equ = astro.get_apparent_posn(equ, j)
            self.assertAlmostEqual(equ.ra, next(ira_out), ACCURACY_RA)
            self.assertAlmostEqual(equ.dec, next(idec_out), ACCURACY_DEC)
            
    def test_B1950_to_J2000(self):
        """Test astro.B1950_to_J2000() function."""
        
        ACCURACY_RA = 2
        ACCURACY_DEC = 2
        
        # Test set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        pos_J2000 = astro.equ_posn(296.39687500, 78.61902778)
        pos_B1950 = astro.equ_posn(296.98456518, 78.49400140)
        
        equ = astro.B1950_to_J2000(pos_B1950)
        self.assertAlmostEqual(equ.ra,  pos_J2000.ra,  ACCURACY_RA)
        self.assertAlmostEqual(equ.dec, pos_J2000.dec, ACCURACY_DEC)
        
    def test_J2000_to_B1950(self):
        """Test astro.J2000_to_B1950() function."""
        
        ACCURACY_RA = 2
        ACCURACY_DEC = 2
        
        # Test set generated at http://ned.ipac.caltech.edu/forms/calculator.html
        pos_J2000 = astro.equ_posn(296.39687500, 78.61902778)
        pos_B1950 = astro.equ_posn(296.98456518, 78.49400140)
        
        equ = astro.J2000_to_B1950(pos_J2000)
        self.assertAlmostEqual(equ.ra,  pos_B1950.ra,  ACCURACY_RA)
        self.assertAlmostEqual(equ.dec, pos_B1950.dec, ACCURACY_DEC)


class astro_test_suite(unittest.TestSuite):
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(astro_tests))
                

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function

"""
Module for analyzing images.  Currently, this module supports:
  * estimating the position-dependent background level in the image
  * finding point sources

.. versionadded:: 1.1.0
"""

import math
import numpy
from scipy.signal import convolve, medfilt
from scipy.interpolate import bisplrep, bisplev

__version__ = "0.1"
__revision__ = "$Rev: 2454 $"
__all__ = ['estimateBackground', 'findPointSources', '__version__', '__revision__', '__all__']


def estimateBackground(image, window=32):
    """
    Given a 2-D image, estimate and return the background a la SExtractor.
    
    This works by:
      1.  Dividing the image into a number of half-overlapped tiles that are
          'window' by 'window' pixels in size.
      2.  Computing the mean and standard deviation within the tile both
          with and without 3*sigma clipping applied.
      3.  Using the tile statistics, determine if the tile is empty, i.e.,
          the standard deviation has changed by less than 20%, or full.  
            a. If the tile is empty, use the clipped mean for the background.
            b. Otherwise, use 2.5*median - 1.5*mean.
      4.  Once all of the tiles have been processes, median filter them to 
          remove those that are anomalously high.
      5.  Build a bicubic spline to interpolate the background back to the
          original image size.
      6.  Evaluate the spline and return.
    
    For more information on the SExtractor method, see:
      * Bertin & Arnouts (1996, AApS, 317, 393)
      * https://www.astromatic.net/pubsvn/software/sextractor/trunk/doc/sextractor.pdf
      * http://www.astr.tohoku.ac.jp/~akhlaghi/sextractor_notes.html
    """
    
    # Figure out how large the background grid should be assuming a half-
    # tile overlap
    nX = 2 * image.shape[0] / window
    nY = 2 * image.shape[1] / window
    
    # Process the background grid
    backgroundX = numpy.zeros((nX,nY), dtype=numpy.float64)
    backgroundY = numpy.zeros((nX,nY), dtype=numpy.float64)
    backgroundBasis = numpy.zeros((nX,nY), dtype=image.dtype)
    for i in xrange(nX):
        for j in xrange(nY):
            ## Extract the tile area
            tile = image[i*window/2:(i+1)*window/2,j*window/2:(j+1)*window/2]
            
            ## Compute the mean and standard deviation - both with and
            ## without progressive clipping
            m0, s0 = numpy.mean(tile), numpy.std(tile)
            valid = numpy.where( numpy.abs(tile-m0) < 3*s0 )
            m1, s1 = numpy.mean(tile[valid]), numpy.std(tile[valid])
            
            ## If the standard deviation hasn't changed by more than 20% use
            ## it's an uncrowded field and we can use the clipped mean.  Other-
            ## wise estimate the mode.
            if abs(s1-s0)/s0 <= 0.2:
                backgroundBasis[i,j] = m1
            else:
                ### This is SExtractor specific and is a correction for the 
                ### clipped distributions that are being used in this step.
                backgroundBasis[i,j] = 2.5*numpy.median(tile[valid]) - 1.5*m1
                
            ## Save the center of this tile for interpolation
            backgroundX[i,j] = i*window/2 + window/4.0
            backgroundY[i,j] = j*window/2 + window/4.0
            
    # Deal with NaNs in the image
    good = numpy.where( numpy.isfinite(backgroundBasis) )
    bad = numpy.where( ~numpy.isfinite(backgroundBasis) )
    backgroundBasis[bad] = numpy.median(backgroundBasis[good])
    
    # Use a median filter to get rid of unusually high regions
    backgroundBasis = medfilt(backgroundBasis, 3)
    
    # Build up the interpolation function
    backgroundX, backgroundY = backgroundX.ravel(), backgroundY.ravel()
    backgroundBasis = backgroundBasis.ravel()
    backgroundInterp = bisplrep(backgroundX, backgroundY, backgroundBasis, 
                            xb=0, xe=image.shape[0], yb=0, ye=image.shape[1], kx=3, ky=3)
        
    # Evaluate
    x, y = numpy.arange(image.shape[0]), numpy.arange(image.shape[1])
    background = bisplev(x, y, backgroundInterp)
    
    # Done
    return background


def findPointSources(image, threshold=4.0, fwhm=1.0, sharp=[0.2,1.0], round=[-1.0,1.0], backgroundSize=16, verbose=True):
    """
    Given a 2-D image, find all of the point sources in it that meet the
    selection criteria provided via the keywords.  These are:
      * threshold: detection threshold in counts about the background
      * fwhm: source full width half max. in pixel
      * sharp: two-element array that defines the lower and upper bounds
               on the source sharpness
      * round: two-element array that defines the lower and upper bounds
               on the source roundness
        
    Background estimation and removal is handled by the estimateBackground()
    function in this module that implements a SExtractor-like method.  This
    can be disabled by setting the 'backgroundSize' keyword to 0.  For 
    details see :func:`lsl.imaging.analysis.estimateBackground`.
        
    The output of this function is a five-element tuple of 1-D NumPy arrays
    that store information for each source.  The elements are:
      * x: intensity-weighted center - x coordinate
      * y: intensity-weighted center - y coordinate
      * flux: peak count value - input image units
      * sharpness: sharpness statistic
      * roundness: roundness statistic
    
    This function is based on the FIND procedure from the AstroIDL User's
    Library that was written by W. Landsman (Hughes STX).
    
    For additional information about the original IDL routines, see:
    http://idlastro.gsfc.nasa.gov/contents.html#C2
    """
    
    # Maximum size of convolution box in pixels 
    maxbox = 13
    
    # Make sure that the FWHM is not too small and compute the corresponding 
    # value for sigma
    if fwhm < 1.0:
        fwhm = 1.0
    sigma = fwhm / (2*math.sqrt(2*math.log(2)))
    
    # Setup the source detection boxes
    ## The detection radius is the larger of 1.5 sigma or 2 pixels
    radius = max([1.5*sigma, 2.0])			# Radius is the
    ## The detection box size is the smaller of the radius or half the
    ## maximum size of the convolution box
    nHalf = min([int(radius),  (maxbox-1)/2])   	
    ## The size of the convolution box
    nBox = 2*nHalf + 1
    
    # Remove the diffuse background.
    if backgroundSize > 0:
        background = estimateBackground(image, backgroundSize)
        cleanImage = image - background
    else:
        cleanImage = image
        
    # Create the raw convolution kernel and a mask that excludes pixels beyond
    # the detection radius
    ## Coordinate setup
    row = numpy.arange(nBox, dtype=numpy.float64) - nHalf
    rx, ry = numpy.meshgrid(row, row)
    ## Build the mask
    mask = (rx**2 + ry**2) <= radius**2
    valid = numpy.where( mask )
    pixels = int(mask.sum())
    ## Build the kernel
    kernel = numpy.exp(-(rx**2 + ry**2)/2.0/sigma**2)
    ## Trim the kernel to take into account the mask and normalize it
    kernel *= mask  
    kernel /= kernel[valid].sum()
    
    # Build a cross-section of the raw kernel for use in computing the 
    # roundness
    c1 = numpy.exp(-row**2/2.0/sigma**2)
    c1 /= c1.sum()
    
    # Convolve the image with the trimmed and normalized kernel
    convImage = convolve(cleanImage, kernel, mode='same')
    
    # Clean out the image edges where the convolution is lacking
    minConvValue = convImage[nHalf:-nHalf,nHalf:-nHalf].min()
    convImage[:nHalf,:] = minConvValue
    convImage[-nHalf:,:] = minConvValue
    convImage[:,:nHalf] = minConvValue
    convImage[:,-nHalf:] = minConvValue
    
    # Exclude the central pixel from the mask and create the arrays that store
    # the relative pixel coordinates for the detection box
    mask[nHalf,nHalf] = 0
    pixels -= 1
    xx, yy = numpy.where(mask)
    xx, yy = xx - nHalf, yy - nHalf
    
    # Find all of the pixels above the threshold and run a simple "de-duplication"
    # to get the maximum inside the detection box
    ix, iy = numpy.where( convImage >= threshold )
    for i in xrange(pixels):
        ox = ix + xx[i]
        oy = iy + yy[i]
        stars, = numpy.where( convImage[ix,iy] >= convImage[ox,oy] )
        ix, iy = ix[stars], iy[stars]
    nGood = len(ix)       
    
    # Setup the output arrays
    ## Position
    x = numpy.zeros(nGood)
    y = numpy.zeros(nGood)
    ## Peak flux
    flux = numpy.zeros(nGood)
    ## Shape statistics
    sharpness = numpy.zeros(nGood)
    roundness = numpy.zeros(nGood)
    
    # Loop over the source positions and compute the various statistics
    nStar = 0
    bad = {'round': 0, 'sharp': 0}
    for i in xrange(nGood): 
        ## Sources are valid until they are not
        validSharpness = True
        validRoundness = True
    
        ## Extract a "postage" stamp of the source and get it's peak
        temp = image[ix[i]-nHalf:ix[i]+nHalf+1,iy[i]-nHalf:iy[i]+nHalf+1]
        cPeak = image[ix[i],iy[i]]			# "d" is actual pixel intensity        
        
        ## Compute the sharpness statistic
        cSharpness = (temp[nHalf,nHalf] - ((mask*temp).sum())/pixels)/cPeak
        if cSharpness < sharp[0] or cSharpness > sharp[1]:
            validSharpness = False
            bad['sharp'] += 1
            
        ## Compute the roundness statistic
        dx = (temp.sum(axis=1)*c1).sum()   
        dy = (temp.sum(axis=0)*c1).sum()
        cRoundness = (dx-dy) / (dx+dy)		# Roundness statistic
        if (dx <= 0 or dy <= 0) or (cRoundness < round[0] or cRoundness > round[1]):
            validRoundness = False
            bad['round'] += 1
            
        ## Flux-weighted center
        xcen = ix[i] + (temp*rx).sum()/temp.sum()
        ycen = iy[i] + (temp*ry).sum()/temp.sum()
        
        ## Save the results if they are valid
        if validSharpness and validRoundness:
            x[nStar] = xcen
            y[nStar] = ycen
            flux[nStar] = cPeak
            sharpness[nStar] = cSharpness
            roundness[nStar] = cRoundness
            nStar += 1
            
        ## Print out information, if requested
        if verbose:
            print("Source #%i" % (i+1,))
            print("  Center:  %.3f, %.3f" % (xcen, ycen))
            print("  Peak: %.3f" % cPeak)
            print("  Sharpness: %.3f%s" % (cSharpness, " (rejected)" if not validSharpness else ""))
            print("  Roundness: %.3f%s" % (cRoundness, " (rejected)" if not validRoundness else ""))
            
    # Print out a summary, if requested
    if verbose:
        print(" ")
        print("Summary")
        print("  Detections: %i" % nGood)
        print("  Valid Detections: %i" % nStar)
        print("  Number Rejected:")
        print("    Sharpness: %i" % bad['sharp'])
        print("    Roundness: %i" % bad['round'])
        
    # Trim the output arrays for the actual number of stars found
    x = x[:nStar]
    y = y[:nStar]
    flux = flux[:nStar]
    sharpness = sharpness[:nStar]
    roundness = roundness[:nStar]
    
    # Done
    return x, y, flux, sharpness, roundness


# -*- coding: utf-8 -*-

"""
Module for computing spectral kurtosis both for instantaneous PSDs and spectrometer 
output.  This module also provides functions to estimate the spectral kurtosis
limits for a given confidence interval in sigma.

This module is based on:
  * Nita & Gary (2010, PASP 155, 595)
  * Nita & Gary (2010, MNRAS 406, L60)
"""

import numpy

from scipy.special import erf
from scipy.stats import betaprime

__version__ = '0.1'
__revision__ = '$Rev: 2389 $'
__all__ = ['mean', 'std', 'var', 'skew', 'getLimits', 'spectralFFT', 'spectralPower', 
        '__version__', '__revision__', '__all__']


def mean(M, N=1):
    """
    Return the expected mean spectral kurtosis value for M points each composed
    of N measurements.
    """
    
    return 1.0


def std(M, N=1):
    """
    Return the expected standard deviation of the spectral kurtosis for M points 
    each composed of N measurements.
    """
    
    return numpy.sqrt( var(M, N) )


def var(M, N=1):
    """
    Return the expected variance (second central moment) of the spectral kurtosis 
    for M points each composed of N measurements.
    """

    return 2.0*N*(N+1)*M**2/ float( (M-1)*(M*N+3)*(M*N+2) )


def skew(M, N=1):
    """
    Return the expected skewness (third central moment) of the spectral kurtosis 
    for M points each composed of N measurements.
    """
    
    m2 = var(M, N)
    
    return 4.0*m2*M / float( (M-1)*(M*N+5)*(M*N+4) ) * ( (N+4)*M*N - 5*N - 2 )


def _alpha(M, N):
    """
    Determine the value of alpha needed to reproduce the spectral kurtosis PDF via a
    Pearson Type VI distribution (betaprime in scipy.stats world).  
    
    .. note::
        This corresponds to Equation (14) in Nita & Gary (2010, MNRAS 406, L60)
    """
    
    m2 = var(M, N)
    m3 = skew(M, N)

    return 1.0/m3**3*(32*m2**5 - 4*m3*m2**3 + 8*m3**2*m2**2 + m3**2*m2 - m3**3 + (8*m2**3 - m3*m2 + m3**2)*numpy.sqrt( 16*m2**4 + 4*m3**2*m2 + m3**2 ))


def _beta(M, N):
    """
    Determine the value of beta needed to reproduce the spectral kurtosis PDF via a
    Pearson Type VI distribution (betaprime in scipy.stats world).
    
    .. note::
        This corresponds to Equation (14) in Nita & Gary (2010, MNRAS 406, L60)
    """
    
    m2 = var(M, N)
    m3 = skew(M, N)

    return 3 + 2*m2/m3**2*(4*m2**2 + numpy.sqrt( 16*m2**4 + 4*m3**2*m2 + m3**2 ))


def _delta(M, N):
    """
    Determine the value of delta needed to shift a Pearson Type VI distribution (betaprime 
    in scipy.stats world) to a mean value of 1.0
    """
    
    a = _alpha(M, N)
    b = _beta(M, N)

    return (b - a - 1) / (b - 1)


def getLimits(sigma, M, N=1):
    """
    Return the limits on the spectral kurtosis value to exclude the specified confidence
    interval in sigma using a Pearson Type VI distribution (betaprime in scipy.stats world).
    The return value is a two-element tuple of lower limit, upper limit.
    
    .. note::
        This corresponds to Section 3.1 in Nita & Gary (2010, MNRAS 406, L60)
    """
    
    # Adjust the NumPy error levels so that we know when ppf() may be suspect
    errLevels = numpy.geterr()
    numpy.seterr(all='raise')
    
    # Convert the sigma to a fraction for the high and low clip levels
    percentClip = ( 1.0 - erf(sigma/numpy.sqrt(2)) ) / 2.0
    
    # Build the Pearson type VI distribution function - parameters then instance
    a = _alpha(M, N)
    b = _beta(M, N)
    d = _delta(M, N)
    
    rv = betaprime(a, b, loc=d)
    
    # Try to get a realistic lower limit, giving up if we hit an overflow error
    try:
        lower = rv.ppf(percentClip)
    except FloatingPointError:
        lower = mean(M, N) - sigma*std(M, N)
        
    # Try to get a realistic upper limit, giving up if we hit an overflow error
    try:
        upper = rv.ppf(1.0-percentClip)
    except FloatingPointError:
        upper = mean(M, N) + sigma*std(M, N)
        
    # Restore the NumPy error levels
    numpy.seterr(**errLevels)
    
    return lower, upper


def spectralFFT(x):
    """
    Compute the spectral kurtosis for a set of unaveraged FFT measurements.
    For a distribution consistent with Gaussian noise, this value should 
    be ~1.
    """
    
    # Convert to power
    xPrime = numpy.abs(x)**2
    
    return spectralPower(xPrime, N=1)


def spectralPower(x, N=1):
    """
    Compute the spectral kurtosis for a set of power measurements averaged over 
    N FFT windows.  For a distribution consistent with Gaussian noise, this value 
    should be ~1.
    """
    
    M = len(x)
    
    k = M*(x**2).sum()/(x.sum())**2 - 1.0
    k *= (M*N+1)/(M-1)
    
    return k